from django.contrib.sites.models import (
    Site,
)
from django.core.management.base import (
    BaseCommand,
)


from ...models import (
    SiteProfile,
)


class Command(BaseCommand):
    help = 'Creates a SiteProfile instance for each instance in the Django sites framework'

    def handle(self, *args, **options):
        sites = Site.objects.all()
        for site in sites:
            site_profile, created = SiteProfile.objects.get_or_create(
                site_ptr=site,
            )
            site_profile.domain = site.domain
            site_profile.name = site.name
            site_profile.save()



from datetime import (
    timedelta,
)


from django.core.management.base import (
    BaseCommand,
)
from django.utils import (
    timezone,
)


from ...models import (
    Subscriber,
)


class Command(BaseCommand):
    help = 'Deletes subscribers who do not have any subscriptions'

    def handle(self, *args, **options):
        one_day_ago = timezone.now() - timedelta(days=1)
        Subscriber.objects.filter(
            subscriptions=None,
        ).filter(
            created__lte=one_day_ago,
        ).delete()



from datetime import (
    datetime,
    timedelta,
    timezone,
)
from io import (
    BytesIO,
)


from django.contrib.auth.models import (
    Permission,
    User,
)
from django.contrib.sessions.middleware import (
    SessionMiddleware,
)
from django.contrib.sites.models import (
    Site,
)
from django.core import (
    mail,
)
from django.core.exceptions import (
    ObjectDoesNotExist,
    PermissionDenied,
)
from django.core.files.uploadedfile import (
    SimpleUploadedFile,
)
from django.core.management import (
    call_command,
)
from django.http import (
    Http404,
)
from django.test import (
    RequestFactory,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone as django_timezone,
)


from PIL import (
    Image,
)


from django_simple_bulk_emailer.models import (
    BulkEmail,
    EmailDocument,
    EmailImage,
    EmailTracker,
    MonthlyStat,
    SiteProfile,
    Subscriber,
    Subscription,
)


def get_default_site():
    return Site.objects.first()


def get_tracker(subject, first=True):
    tracker_set = EmailTracker.objects.filter(
        subject=subject,
    )
    if first:
        return tracker_set.first()
    else:
        return tracker_set


def get_monthly_stat(year_int, month_int):
    return MonthlyStat.objects.get(
        year_int=year_int,
        month_int=month_int,
    )


def fake_now(year=2020, month=1):
    return datetime(year, month, 1, tzinfo=timezone.utc)


def create_site():
    site_profile = Site.objects.create(
        domain='second.example.com',
        name='Second test site',
    )
    return site_profile


def create_site_profile():
    site_profile = SiteProfile.objects.create(
        protocol='http://',
        domain='127.0.0.1:8000',
        name='Development server',
    )
    return site_profile


def create_subscription(list_name='Test list', publicly_visible=True, mc_sync=False, mc_list='test', sort_order=0):
    subscription = Subscription.objects.create(
        list_name=list_name,
        descriptive_text='Test list description',
        publicly_visible=publicly_visible,
        use_pages=True,
        mc_sync=mc_sync,
        mc_list=mc_list,
        sort_order=sort_order,
    )
    return subscription


def create_subscriber(subscriber_email='example@example.com', first_name='Anonymous', last_name='Subscriber'):
    subscriber = Subscriber.objects.create(
        subscriber_email=subscriber_email,
        first_name=first_name,
        last_name=last_name,
    )
    return subscriber


def create_image_file():
    temp_handle = BytesIO()
    image_file = Image.new(
        'RGB',
        (72, 72),
        (0, 0, 255),
    )
    image_file.save(
        temp_handle,
        'jpeg',
    )
    temp_handle.seek(0)
    return SimpleUploadedFile(
        'test_image.jpeg',
        temp_handle.read(),
        'image/jpeg',
    )


def create_document_file():
    return SimpleUploadedFile(
            'test_file.pdf',
            'test file content'.encode(),
            'application/pdf',
        )


def create_email(headline=None, list_name=None, body_text=None, published=False, sendable=False):
    if not headline:
        headline = 'Test headline of many characters'
    if not body_text:
        body_text = '<p>Test body text paragraph one.</p><p>Test body text paragraph two.</p>'
    if list_name:
        subscription_list = Subscription.objects.filter(
            list_name=list_name,
        ).first()
    else:
        subscription_list = create_subscription()
    bulk_email = BulkEmail.objects.create(
        subscription_list=subscription_list,
        headline=headline,
        body_text=body_text,
        published=published,
        sendable=sendable,
    )
    EmailImage.objects.create(
        bulk_email=bulk_email,
        caption='Test caption',
        description='Test description',
        image_width=1080,
        saved_file=create_image_file(),
    )
    EmailDocument.objects.create(
        bulk_email=bulk_email,
        title='Test title',
        extra_text='Test extra text',
        saved_file=create_document_file()
    )
    return bulk_email


def create_tracker(subject='Test email subject', subscription_name='Test subscription', send_complete=django_timezone.now(), number_sent=0, json_data=None):
    if json_data is None:
        json_data = {}
    tracker = EmailTracker.objects.create(
        subject=subject,
        subscription_name=subscription_name,
        send_complete=send_complete,
        number_sent=number_sent,
        json_data=json_data,
    )
    return tracker


def create_stats(year_int=2020, month_int=1, stat_data='Test stat data'):
    stats = MonthlyStat.objects.create(
        year_int=year_int,
        month_int=month_int,
        stat_data=stat_data,
    )
    return stats


def create_user(permission_list=None):
    user = User.objects.create_user(
        username='test_user',
        password='test_password',
    )
    if permission_list:
        user.is_staff = True
        user.save()
        for permission in permission_list:
            added_permission = Permission.objects.get(
                codename=permission,
            )
            user.user_permissions.add(added_permission)
    return user


def dummy_get_response(request):
    return None


def create_request_response(self, get_post, page='1', key='empty', time_dict=None):
    reverse_name = f'django_simple_bulk_emailer:{self.view_name}'
    with self.settings(
        SITE_ID=self.profile_instance.site_ptr.id,
    ):
        reverse_string = reverse(
            reverse_name,
            kwargs=self.kwargs,
        )
        if page != '1':
            reverse_string = f'{reverse_string}?page={page}'
        if key != 'empty':
            reverse_string = f'{reverse_string}?key={key}'
        if get_post == 'post':
            self.request = RequestFactory().post(
                reverse_string,
                self.data,
            )
        else:
            self.request = RequestFactory().get(
                reverse_string,
            )
        try:
            self.request.user = self.user
        except AttributeError:
            pass
        SessionMiddleware(dummy_get_response).process_request(self.request)
        if time_dict:
            load_time = django_timezone.now() - timedelta(**time_dict)
            self.request.session['form_load_time'] = load_time.timestamp()
            self.request.session.save()
        args = []
        for key, value in self.kwargs.items():
            args.append(value)
        self.response = self.test_view(
            self.request,
            *args,
        )


def create_subscriber_subscription_state(self, subscriber_email, first_name, last_name, state):
    """ """
    '''
    0: Does not exist
    1: Exists with no subscriptions
    2: Exists with subscription one
    3: Exists with subscription two
    4: Exists with subscriptions one and two
    '''
    if state != 0:
        subscriber = create_subscriber(
            subscriber_email=subscriber_email,
            first_name=first_name,
            last_name=last_name,
        )
        if state == 2 or state == 4:
            try:
                subscriber.subscriptions.add(self.subscription_one)
            except AttributeError:
                pass
        if state == 3 or state == 4:
            try:
                subscriber.subscriptions.add(self.subscription_two)
            except AttributeError:
                pass
        return subscriber


def create_subscription_email_state(self, list_name, state):
    """ """
    '''
    0: 'does not exist',
    1: 'exists with no emails',
    2: 'exists with one email, not sendable',
    3: 'exists with one email, sendable',
    4: 'exists with two emails, none sendable',
    5: 'exists with two emails, first sendable',
    6: 'exists with two emails, second sendable',
    7: 'exists with two emails, both sendable',
    '''
    if state != 0:
        subscription = create_subscription(
            list_name=list_name,
        )
        if list_name == self.list_one_name:
            self.subscription_one = subscription
        if list_name == self.list_two_name:
            self.subscription_two = subscription
        if state != 1:
            email_one = create_email(
                list_name=list_name,
            )
        if state > 3:
            email_two = create_email(
                list_name=list_name,
            )
        if state == 3 or state == 5 or state == 7:
            email_one.sendable = True
            email_one.save()
        if state > 5:
            email_two.sendable = True
            email_two.save()


def call_test_command(self):
    try:
        with self.settings(
                SITE_ID=self.profile_instance.site_ptr.id,
        ):
            call_command(
                self.test_command,
            )
    except AttributeError:
        call_command(
            self.test_command,
        )


def remove_subscriber(subscriber_email):
    try:
        subscriber = Subscriber.objects.get(
            subscriber_email=subscriber_email,
        )
        for subscription in subscriber.subscriptions.all():
            subscriber.subscriptions.remove(subscription)
        subscriber.delete()
    except ObjectDoesNotExist:
        pass


def remove_subscription_and_emails(list_name):
    try:
        subscription = Subscription.objects.get(
            list_name=list_name,
        )
        BulkEmail.objects.filter(
            subscription_list=subscription,
        ).delete()
        subscription.delete()
    except ObjectDoesNotExist:
        pass


def clear_data_and_files():
    BulkEmail.objects.all().delete()
    EmailTracker.objects.all().delete()
    MonthlyStat.objects.all().delete()
    SiteProfile.objects.all().delete()
    Subscription.objects.all().delete()
    Subscriber.objects.all().delete()


def attribute_equals(self, attr_dict, command=False):
    if command:
        inserted_text = f"command '{self.test_command}'"
    else:
        inserted_text = f"model '{self.test_instance.__class__.__name__}'"
    for key, value in attr_dict.items():
        error_msg = f"For {inserted_text}, the value of attribute '{key}' was not '{value}'"
        attr = getattr(self.test_instance, key)
        self.assertTrue(attr == value, error_msg)


def attribute_length_equals(self, attr_name, value):
    error_msg = f"For model '{self.test_instance.__class__.__name__}', the length of attribute '{attr_name}' was not '{value}'"
    attr = getattr(self.test_instance, attr_name)
    self.assertIs(len(attr), value, error_msg)


def method_output_equals(self, attr_name, value):
    error_msg = f"For model '{self.test_instance.__class__.__name__}', the output of method '{attr_name}' was not '{value}'"
    function = getattr(self.test_instance, attr_name)
    self.assertTrue(function() == value, error_msg)


def method_output_contains(self, attr_name, value):
    error_msg = f"For model '{self.test_instance.__class__.__name__}', the output of method '{attr_name}' did not contain '{value}'"
    function = getattr(self.test_instance, attr_name)
    self.assertIn(value, function(), error_msg)


def method_output_length_equals(self, attr_name, value):
    error_msg = f"For model '{self.test_instance.__class__.__name__}', the length of the output of method '{attr_name}' was not '{value}'"
    function = getattr(self.test_instance, attr_name)
    self.assertIs(len(function()), value, error_msg)


def email_exists(self, headline, boolean):
    error_msg_true = f"An email with headline '{headline}' does not exist"
    error_msg_false = f"An email with headline '{headline}' should not exist"
    try:
        bulk_email = BulkEmail.objects.get(
            headline=headline,
        )
    except ObjectDoesNotExist:
        bulk_email = BulkEmail.objects.none()
    if boolean:
        self.assertTrue(bulk_email, error_msg_true)
    else:
        self.assertFalse(bulk_email, error_msg_false)


def stats_exist(self, year_int, month_int, boolean):
    error_msg_true = f"Monthly stats with year '{str(year_int)}' and month '{str(month_int)}' do not exist"
    error_msg_false = f"Monthly stats with year '{str(year_int)}' and month '{str(month_int)}' should not exist"
    try:
        stats = MonthlyStat.objects.get(
            year_int=year_int,
            month_int=month_int,
        )
    except ObjectDoesNotExist:
        stats = MonthlyStat.objects.none()
    if boolean:
        self.assertTrue(stats, error_msg_true)
    else:
        self.assertFalse(stats, error_msg_false)


def subscriber_exists(self, subscriber_email, boolean, extra_text=''):
    try:
        first_text = f"For view '{self.view_name}', a"
    except AttributeError:
        first_text = 'A'
    error_msg_true = f"{first_text} subscriber with email address '{subscriber_email}' does not exist{extra_text}"
    error_msg_false = f"{first_text} subscriber with email address '{subscriber_email}' should not exist{extra_text}"
    try:
        subscriber = Subscriber.objects.get(
            subscriber_email=subscriber_email,
        )
    except ObjectDoesNotExist:
        subscriber = Subscriber.objects.none()
    if boolean:
        self.assertTrue(subscriber, error_msg_true)
    else:
        self.assertFalse(subscriber, error_msg_false)


def check_subscriber_count(self, count):
    subscriber_count = Subscriber.objects.count()
    error_msg = f"For view '{self.view_name}', the number of subscribers was '{str(subscriber_count)}', not '{str(count)}'"
    self.assertEqual(subscriber_count, count, error_msg)


def check_subscription_count(self, count):
    subscription_count = self.subscriber.subscriptions.count()
    error_msg = f"For view '{self.view_name}', subscriber '{self.subscriber.subscriber_email}', the number of subscriptions was '{str(subscription_count)}', not '{str(count)}'"
    self.assertEqual(subscription_count, count, error_msg)


def check_attached_trackers(self, count_dict=None, set_dict=None):
    if count_dict is None:
        count_dict = {}
    if set_dict is None:
        set_dict = {}
    for key, value in count_dict.items():
        tracker_count = getattr(self.test_instance, key).count()
        error_msg = f"For command '{self.test_command}', the number of trackers assigned to '{key}' was '{str(tracker_count)}', not '{str(value)}'"
        self.assertEqual(tracker_count, value, error_msg)
    for key, value in set_dict.items():
        tracker_list = list(getattr(self.test_instance, key).all())
        value_list = list(value)
        error_msg = f"For command '{self.test_command}', the list of trackers assigned to '{key}' was '{str(tracker_list)}', not '{str(value_list)}'"
        self.assertEqual(tracker_list, value_list, error_msg)


def check_site_profile_count(self, count):
    site_profile_count = SiteProfile.objects.count()
    error_msg = f"The number of site profiles was '{str(site_profile_count)}', not '{str(count)}'"
    self.assertEqual(site_profile_count, count, error_msg)


def check_site_profile(self, test_site, test_dict):
    error_msg_exists = f"A site profile matching the site with domain '{test_dict['domain']}' does not exist"
    try:
        site_profile = SiteProfile.objects.get(
            site_ptr=test_site,
        )
        profile_created = True
        for key, value in test_dict.items():
            attr_value = getattr(site_profile, key)
            error_msg_attr = f"A site profile that should have had the value '{test_dict[key]}' for attribute '{key}' instead had '{attr_value}'"
            self.assertEqual(value, attr_value, error_msg_attr)
        error_msg_protocol = f"The site profile with domain '{site_profile.domain}' had the protocol '{site_profile.protocol}' and not 'https://'"
        self.assertEqual(site_profile.protocol, 'https://', error_msg_protocol)
    except ObjectDoesNotExist:
        profile_created = False
    self.assertTrue(profile_created, error_msg_exists)


def check_subscriber_attributes(self, subscriber_email, attributes, boolean, extra_text=''):
    checked_subscriber = Subscriber.objects.get(
        subscriber_email=subscriber_email,
    )
    subscription_pks = checked_subscriber.subscriptions.values_list('pk', flat=True)
    subscription_list = []
    for pk in subscription_pks:
        subscription_list.append(str(pk))
    subscriber_data = {
        'first_name': checked_subscriber.first_name,
        'last_name': checked_subscriber.last_name,
        'subscriber_email': checked_subscriber.subscriber_email,
        'subscription_choices': subscription_list,
        'mc_email': checked_subscriber.mc_email,
    },
    for key in attributes.keys():
        error_msg_true = f"For view '{self.view_name}', subscriber '{subscriber_email}', the value of '{key}' was '{subscriber_data[0][key]}', not '{attributes[key]}'{extra_text}"
        error_msg_false = f"For view '{self.view_name}', subscriber '{subscriber_email}', the value of '{key}' should not be '{attributes[key]}'{extra_text}"
        if boolean:
            self.assertEqual(attributes[key], subscriber_data[0][key], error_msg_true)
        else:
            self.assertNotEqual(attributes[key], subscriber_data[0][key], error_msg_false)


def check_quantity_email_sent(self, quantity, clear_outbox=False, extra_text=''):
    try:
        first_text = f"For view '{self.view_name}', t"
    except AttributeError:
        first_text = 'T'
    emails_sent = len(mail.outbox)
    error_msg = f"{first_text}he number of emails sent was '{str(emails_sent)}', not '{str(quantity)}'{extra_text}"
    self.assertEqual(emails_sent, quantity, error_msg)
    if clear_outbox:
        mail.outbox = []


def check_quantity_trackers(self, quantity):
    tracker_count = EmailTracker.objects.count()
    error_msg = f"For command '{self.test_command}', the number of trackers existing was '{str(tracker_count)}', not '{str(quantity)}'"
    self.assertEqual(tracker_count, quantity, error_msg)


def check_email(self, email_number=0, subject=None, text_strings=None, html_strings=None):
    checked_email = mail.outbox[email_number]
    email_subject = checked_email.subject
    try:
        inserted_text = f"view '{self.view_name}'"
    except AttributeError:
        inserted_text = f"command '{self.test_command}'"
    error_msg_one = f"For {inserted_text}, email number '{str(email_number)}' did not have HTML alternative attached"
    error_msg_two = f"For {inserted_text}, email number '{str(email_number)}', HTML attachment did not have the MIME type 'text/html'"
    error_msg_three = f"For {inserted_text}, email number '{str(email_number)}' had the subject '{email_subject}', not '{subject}'"
    self.assertEqual(len(checked_email.alternatives), 1, error_msg_one)
    self.assertEqual(checked_email.alternatives[0][1], 'text/html', error_msg_two)
    if subject:
        self.assertEqual(email_subject, subject, error_msg_three)
    email_contains(
        self,
        checked_email,
        email_number,
        inserted_text,
        text_strings,
        html_strings,
    )


def status_code_equals(self, status_code):
    returned_code = self.response.status_code
    error_msg = f"For view '{self.view_name}', the status code returned was '{str(returned_code)}', not '{str(status_code)}'"
    self.assertEqual(returned_code, status_code, error_msg)


def redirect_url_equals(self, redirect_url):
    returned_url = self.response.url
    error_msg = f"For view '{self.view_name}', the redirect URL returned was '{returned_url}', not '{redirect_url}'"
    self.assertEqual(returned_url, redirect_url, error_msg)


def session_contains(self, value, boolean):
    error_msg_true = f"For view '{self.view_name}', the session did not contain '{value}'"
    error_msg_false = f"For view '{self.view_name}', the session should not contain '{value}'"
    if boolean:
        self.assertTrue(value in self.request.session, error_msg_true)
    else:
        self.assertFalse(value in self.request.session, error_msg_false)


def email_contains(self, checked_email, email_number, inserted_text, text_strings=None, html_strings=None):
    if text_strings is None:
        text_strings = []
    if html_strings is None:
        html_strings = []
    for string in text_strings:
        error_msg = f"For {inserted_text}, email number '{str(email_number)}' did not include '{string}' in its text body"
        self.assertIn(string, checked_email.body, error_msg)
    for string in html_strings:
        error_msg_one = f"For {inserted_text}, email number '{str(email_number)}' should not include '{string}' in its text body"
        error_msg_two = f"For {inserted_text}, email number '{str(email_number)}' did not include '{string}' in its HTML body"
        self.assertNotIn(string, checked_email.body, error_msg_one)
        self.assertIn(string, checked_email.alternatives[0][0], error_msg_two)


def html_contains(self, test_string, true_strings=None, false_strings=None):
    if true_strings is None:
        true_strings = []
    if false_strings is None:
        false_strings = []
    try:
        inserted_text = f"view '{self.view_name}'"
    except AttributeError:
        inserted_text = f"command '{self.test_command}'"
    for string in true_strings:
        error_msg = f"For {inserted_text}, the HTML did not contain '{string}'"
        self.assertTrue(string in test_string, error_msg)
    for string in false_strings:
        error_msg = f"For {inserted_text}, the HTML should not contain '{string}'"
        self.assertFalse(string in test_string, error_msg)


def json_contains(self, json_data=None, true_dict=None, false_dict=None):
    if true_dict is None:
        true_dict = {}
    if false_dict is None:
        false_dict = {}
    if json_data is None:
        self.tracker.refresh_from_db()
        data = self.tracker.json_data
    else:
        data = json_data
    for key, value in true_dict.items():
        error_msg_key = f"For view '{self.view_name}', the JSON data did not contain the key '{str(key)}'"
        error_msg_value = f"For view '{self.view_name}', the JSON data value for key '{str(key)}' was '{str(data[key])}' and not '{str(value)}'"
        self.assertIn(key, data.keys(), error_msg_key)
        self.assertEqual(value, data[key], error_msg_value)

    for key, value in false_dict.items():
        error_msg_key = f"For view '{self.view_name}', the JSON data should not contain the key '{str(key)}'"
        error_msg_value = f"For view '{self.view_name}', the JSON data value for key '{str(key)}' should not be '{str(value)}'"
        self.assertNotIn(key, data.keys(), error_msg_key)
        self.assertFalse(value, data[key], error_msg_value)


def image_contains(self, image_dict=None):
    if image_dict is None:
        image_dict = {}
    image_data = BytesIO(self.response.content)
    image_file = Image.open(image_data)
    for key, value in image_dict.items():
        error_msg = f"For view '{self.view_name}', the returned image's property '{key}' was not '{str(value)}'"
        image_property = getattr(image_file, key)
        self.assertIs(image_property, value, error_msg)


def check_http_response(self, form_load=False, status_code=200, redirect_url=None, true_strings=None, false_strings=None, true_dict=None, false_dict=None, image_dict=None):
    status_code_equals(
        self,
        status_code,
    )
    if redirect_url:
        redirect_url_equals(
            self,
            redirect_url,
        )
    session_contains(
        self,
        'form_load_time',
        form_load,
    )
    if true_strings or false_strings:
        html_contains(
            self,
            self.response.content.decode(),
            true_strings=true_strings,
            false_strings=false_strings,
        )
    if true_dict or false_dict:
        json_contains(
            self,
            true_dict=true_dict,
            false_dict=false_dict,
        )
    if image_dict:
        image_contains(
            self,
            image_dict,
        )


def check_permission(self, boolean):
    try:
        create_request_response(
            self,
            'get',
        )
        permission = True
    except PermissionDenied:
        permission = False
    error_msg_true = f"For view '{self.view_name}', access permission was denied"
    error_msg_false = f"For view '{self.view_name}', access permission was given"
    if boolean:
        self.assertTrue(permission, error_msg_true)
    else:
        self.assertFalse(permission, error_msg_false)


def check_not_found(self, boolean):
    try:
        create_request_response(
            self,
            'get',
        )
        not_found = False
    except Http404:
        not_found = True
    error_msg_true = f"For view '{self.view_name}', a 404 error was not returned"
    error_msg_false = f"For view '{self.view_name}', a 404 error was returned"
    if boolean:
        self.assertTrue(not_found, error_msg_true)
    else:
        self.assertFalse(not_found, error_msg_false)


def check_subscriber_subscription_state(self, subscriber_email, attributes, state, extra_text=''):
    """ """
    '''
    States
    0: Does not exist
    1: Exists with no subscriptions
    2: Exists with subscription one
    3: Exists with subscription two
    4: Exists with subscriptions one and two
    '''
    if state == 0:
        subscriber_exists(
            self,
            subscriber_email,
            False,
            extra_text=extra_text,
        )
    else:
        subscriber_exists(
            self,
            subscriber_email,
            True,
            extra_text=extra_text,
        )
        subscription_one = str(self.subscription_one.pk)
        subscription_two = str(self.subscription_two.pk)
        subscription_choices = []
        if state == 2 or state == 4:
            subscription_choices += [
                subscription_one,
            ]
        if state == 3 or state == 4:
            subscription_choices += [
                subscription_two,
            ]
        attributes['subscription_choices'] = subscription_choices
        check_subscriber_attributes(
            self,
            subscriber_email,
            attributes,
            True,
            extra_text=extra_text,
        )


def check_form_is_valid(self, boolean, extra_text=''):
    self.form = self.form_class(
        data=self.data,
    )
    error_msg_true = f'The form is not valid but should be{extra_text}'
    error_msg_false = f'The form is valid but should not be{extra_text}'
    if boolean:
        self.assertTrue(self.form.is_valid(), error_msg_true)
    else:
        self.assertFalse(self.form.is_valid(), error_msg_false)


def check_field_in_form(self, key, boolean):
    self.form = self.form_class(
        data=self.data,
    )
    self.form.save(commit=False)
    if key in self.form.cleaned_data.keys():
        key_present = True
    else:
        key_present = False
    error_msg_true = f"The key '{key}' was not present in the form's cleaned data and should have been"
    error_msg_false = f"The key '{key}' was present in the form's cleaned data and should not have been"
    if boolean:
        self.assertTrue(key_present, error_msg_true)
    else:
        self.assertFalse(key_present, error_msg_false)


def compare_secret_keys(self, secret_key_old, secret_key_new):
    error_msg = 'The secret key was not changed'
    self.assertNotEqual(secret_key_old, secret_key_new, error_msg)



from unittest.mock import (
    patch,
)


from django.test import (
    TestCase,
)
from django.urls import (
    reverse,
)

from django_simple_bulk_emailer.views import (
    email_preview,
    get_subscriptions,
    list_view,
    manage_subscriptions,
    mc_sync,
    opened_email,
    page_preview,
    page_view,
    quick_unsubscribe,
)


from .functions import (
    attribute_equals,
    check_email,
    check_http_response,
    check_not_found,
    check_permission,
    check_quantity_email_sent,
    compare_secret_keys,
    check_subscriber_attributes,
    check_subscriber_count,
    check_subscription_count,
    check_subscriber_subscription_state,
    clear_data_and_files,
    remove_subscriber,
    create_email,
    create_request_response,
    create_site_profile,
    create_subscriber,
    create_subscription,
    create_tracker,
    create_subscriber_subscription_state,
    create_user,
    fake_now,
    json_contains,
    subscriber_exists,
)


class MixinWrap:
    class BaseMixin(TestCase):
        longMessage = False

        def setUp(self):
            self.profile_instance = create_site_profile()
            self.test_view = eval(self.view_name)
            super().setUp()

        def tearDown(self):
            clear_data_and_files()
            super().tearDown()


class GetSubscriptionsTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.view_name = 'get_subscriptions'
        self.kwargs = {
        }
        self.time_dict = {
            'seconds': 1,
        }
        create_subscriber()
        super().setUp()

    def test_get(self):
        create_request_response(
            self,
            'get',
        )
        check_subscriber_count(
            self,
            1,
        )
        check_quantity_email_sent(
            self,
            0,
        )
        check_http_response(
            self,
            form_load=True,
            true_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
            false_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
        )

    def test_post_new_subscriber(self):
        test_address = 'new@example.com'
        self.data = {
            'subscriber_email': test_address,
        }
        create_request_response(
            self,
            'post',
            time_dict=self.time_dict
        )
        check_subscriber_count(
            self,
            2,
        )
        subscriber_exists(
            self,
            test_address,
            True,
        )
        check_quantity_email_sent(
            self,
            1,
        )
        check_email(
            self,
            subject='Manage your email subscriptions',
            text_strings=[
                'You can select your email subscriptions',
            ],
            html_strings=[
                '<!DOCTYPE html>',
            ],
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
            false_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
        )

    def test_post_existing_subscriber(self):
        test_address = 'example@example.com'
        self.data = {
            'subscriber_email': test_address,
        }
        create_request_response(
            self,
            'post',
            time_dict=self.time_dict
        )
        subscriber_exists(
            self,
            test_address,
            True,
        )
        check_subscriber_count(
            self,
            1,
        )
        check_quantity_email_sent(
            self,
            1,
        )
        check_email(
            self,
            subject='Manage your email subscriptions',
            text_strings=[
                'You can select your email subscriptions',
            ],
            html_strings=[
                '<!DOCTYPE html>',
            ],
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
            false_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
        )

    def test_post_invalid_email(self):
        test_address = 'invalid_example.com'
        self.data = {
            'subscriber_email': test_address,
        }
        create_request_response(
            self,
            'post',
            time_dict=self.time_dict
        )
        subscriber_exists(
            self,
            test_address,
            False,
        )
        check_subscriber_count(
            self,
            1,
        )
        check_quantity_email_sent(
            self,
            0,
        )
        check_http_response(
            self,
            form_load=True,
            true_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
            false_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
        )

    def test_post_fast_submit(self):
        test_address = 'fast@example.com'
        self.data = {
            'subscriber_email': test_address,
        }
        create_request_response(
            self,
            'post',
            time_dict={
                'seconds': 0,
            }
        )
        subscriber_exists(
            self,
            test_address,
            False,
        )
        check_subscriber_count(
            self,
            1,
        )
        check_quantity_email_sent(
            self,
            0,
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
            false_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
        )

    def test_post_honeypot_content(self):
        test_address = 'honeypot@example.com'
        self.data = {
            'subscriber_email': test_address,
            'email': test_address,
        }
        create_request_response(
            self,
            'post',
            time_dict=self.time_dict
        )
        subscriber_exists(
            self,
            test_address,
            False,
        )
        check_subscriber_count(
            self,
            1,
        )
        check_quantity_email_sent(
            self,
            0,
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. An email with instructions has been sent to the address provided.',
            ],
            false_strings=[
                'To subscribe or manage your subscriptions, submit your email address.',
            ],
        )


class ManageSubscriptionsTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.subscriber = create_subscriber()
        self.subscription_one = create_subscription(list_name='List One')
        self.subscription_two = create_subscription(list_name='List Two')
        self.subscriber.subscriptions.add(self.subscription_one)
        self.view_name = 'manage_subscriptions'
        self.kwargs = {
            'subscriber_key': self.subscriber.subscriber_key,
        }
        super().setUp()

    def test_get_valid_key(self):
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'First name:',
                'name="first_name" value="Anonymous"',
                'name="last_name" value="Subscriber"',
                'name="subscriber_email" value="example@example.com"',
                'List One',
                'List Two',
                '" checked> List One',
                '"> List Two',
            ],
            false_strings=[
                'The access link is invalid or has expired.',
                '"> List One',
                '" checked> List Two',
            ],
        )

    def test_get_invalid_key(self):
        self.kwargs['subscriber_key'] = 'InvalidSubscriberKey'
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'The access link is invalid or has expired.',
                '/subscriptions/subscribe/">',
            ],
            false_strings=[
                'First name:',
            ],
        )

    def test_post_unsubscribe_all(self):
        self.data = {
            'unsubscribe_all': 'Unsubscribe from all',
        }
        create_request_response(
            self,
            'post',
        )
        check_subscription_count(
            self,
            0,
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. Your changes have been saved.',
            ],
        )

    def test_post_valid_email(self):
        test_email = 'new@example.com'
        self.data = {
            'first_name': 'Updated',
            'last_name': 'Name',
            'subscriber_email': test_email,
            'subscription_choices': ['1', '2'],
        }
        create_request_response(
            self,
            'post',
        )
        check_subscription_count(
            self,
            2,
        )
        check_subscriber_attributes(
            self,
            test_email,
            self.data,
            True,
        )
        check_http_response(
            self,
            true_strings=[
                'Thank you. Your changes have been saved.',
            ],
        )

    def test_post_invalid_email(self):
        self.data = {
            'first_name': 'Updated',
            'last_name': 'Name',
            'subscriber_email': 'new_example.com',
            'subscription_choices': ['1', '2'],
        }
        create_request_response(
            self,
            'post',
        )
        check_subscription_count(
            self,
            1,
        )
        check_subscriber_attributes(
            self,
            'example@example.com',
            self.data,
            False,
        )
        check_http_response(
            self,
            true_strings=[
                'First name:',
                'name="first_name" value="Updated"',
                'name="last_name" value="Name"',
                'name="subscriber_email" value="example@example.com"',
                '" checked> List One',
                '" checked> List Two',
            ],
            false_strings=[
                'Thank you. Your changes have been saved.',
                '"> List One',
                '"> List Two',
            ],
        )


class QuickUnsubscribeTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.subscriber = create_subscriber()
        self.subscription_one = create_subscription(list_name='List One')
        self.subscription_two = create_subscription(list_name='List Two')
        self.subscriber.subscriptions.add(self.subscription_one)
        self.view_name = 'quick_unsubscribe'
        self.kwargs = {
            'list_slug': self.subscription_one.list_slug,
            'subscriber_key': self.subscriber.subscriber_key,
        }
        super().setUp()

    def test_get_valid(self):
        create_request_response(
            self,
            'get',
        )
        check_subscription_count(
            self,
            0,
        )
        check_http_response(
            self,
            true_strings=[
                'You have been unsubscribed from the List One email distribution list.',
                '/subscriptions/manage/',
            ],
            false_strings=[
                'The access link is invalid or has expired.',
            ],
        )

    def test_get_invalid_slug(self):
        self.kwargs['list_slug'] = 'InvalidListSlug'
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'The access link is invalid or has expired.',
                '/subscriptions/subscribe/">',
            ],
            false_strings=[
                'First name:',
            ],
        )

    def test_get_invalid_key(self):
        self.kwargs['subscriber_key'] = 'InvalidSubscriberKey'
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'The access link is invalid or has expired.',
                '/subscriptions/subscribe/">',
            ],
            false_strings=[
                'First name:',
            ],
        )


class EmailPreviewTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.bulk_email = create_email()
        self.view_name = 'email_preview'
        self.kwargs = {
            'list_slug': self.bulk_email.subscription_list.list_slug,
            'pk': self.bulk_email.pk,
        }
        super().setUp()

    def test_get_without_permission(self):
        self.user = create_user()
        check_permission(
            self,
            False,
        )

    def test_get_with_view_permission(self):
        self.user = create_user(
            permission_list=[
                'view_bulkemail',
            ],
        )
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'Sending history:',
                'Email has not been sent.',
                'Return to list',
                'Test headline of many characters',
                'img src="http://127.0.0.1:8000/media/images/',
                'width="590"',
                'alt="Test description"',
                'Test caption',
                'Test body text paragraph one.',
                'a href="http://127.0.0.1:8000/media/documents/temporary/test-title-',
                'Test title',
                'Test extra text',
                'a href="/mail_test/test-list/',
                '/test-headline-of-many-characters.html',
            ],
            false_strings=[
                'Send email',
                'Email currently is being sent.',
                'Send email again',
            ],
        )

    def test_get_with_change_permission(self):
        self.user = create_user(
            permission_list=[
                'change_bulkemail',
                'view_bulkemail',
            ],
        )
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'Sending history:',
                'Email has not been sent.',
                'Send email',
                'Return to list',
                'Test headline of many characters',
                'img src="http://127.0.0.1:8000/media/images/',
                'width="590"',
                'alt="Test description"',
                'Test caption',
                'Test body text paragraph one.',
                'a href="http://127.0.0.1:8000/media/documents/temporary/test-title-',
                'Test title',
                'Test extra text',
                'a href="/mail_test/test-list/',
                '/test-headline-of-many-characters.html',
            ],
            false_strings=[
                'Email currently is being sent.',
                'Send email again',
            ],
        )

    def test_get_invalid_list(self):
        self.user = create_user(
            permission_list=[
                'view_bulkemail',
            ],
        )
        self.kwargs = {
            'list_slug': 'invalid-slug',
            'pk': self.bulk_email.pk,
        }
        check_not_found(
            self,
            True,
        )

    def test_get_invalid_key(self):
        self.user = create_user(
            permission_list=[
                'view_bulkemail',
            ],
        )
        self.kwargs = {
            'list_slug': self.bulk_email.subscription_list.list_slug,
            'pk': 999,
        }
        check_not_found(
            self,
            True,
        )

    def test_post_return_list(self):
        self.data = {
            'return_list': 'Return to list',
        }
        self.user = create_user(
            permission_list=[
                'change_bulkemail',
                'view_bulkemail',
            ],
        )
        create_request_response(
            self,
            'post',
        )
        check_http_response(
            self,
            status_code=302,
            redirect_url='/admin/django_simple_bulk_emailer/bulkemail/',
        )

    def test_post_send_email(self):
        self.data = {
            'send_email': 'Send email',
        }
        self.user = create_user(
            permission_list=[
                'change_bulkemail',
                'view_bulkemail',
            ],
        )
        create_request_response(
            self,
            'post',
        )
        check_http_response(
            self,
            status_code=302,
            redirect_url='/admin/django_simple_bulk_emailer/bulkemail/',
        )
        self.bulk_email.refresh_from_db()
        self.test_instance = self.bulk_email
        attribute_equals(
            self,
            {
                'sendable': True,
                'sending': True,
                'sent': True,
            },
        )
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'Sending history:',
                'Email currently is being sent.',
                'Return to list',
                'Test headline of many characters',
                'img src="http://127.0.0.1:8000/media/images/',
                'width="590"',
                'alt="Test description"',
                'Test caption',
                'Test body text paragraph one.',
                'a href="http://127.0.0.1:8000/media/documents/temporary/test-title-',
                'Test title',
                'Test extra text',
                'a href="/mail_test/test-list/',
                '/test-headline-of-many-characters.html',
            ],
            false_strings=[
                'Email has not been sent.',
                'Send email',
                'Send email again',
            ],
        )
        self.bulk_email.sendable = False
        self.bulk_email.send_history = 'Test sending history'
        self.bulk_email.sending = False
        self.bulk_email.save()
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'Sending history:',
                'Test sending history',
                'Send email again',
                'Return to list',
                'Test headline of many characters',
                'img src="http://127.0.0.1:8000/media/images/',
                'width="590"',
                'alt="Test description"',
                'Test caption',
                'Test body text paragraph one.',
                'a href="http://127.0.0.1:8000/media/documents/temporary/test-title-',
                'Test title',
                'Test extra text',
                'a href="/mail_test/test-list/',
                '/test-headline-of-many-characters.html',
            ],
            false_strings=[
                'Email has not been sent.',
                'Email currently is being sent.',
                'Send email<',
            ],
        )


class ListViewTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.subscription = create_subscription()
        self.view_name = 'list_view'
        self.kwargs = {
            'list_slug': self.subscription.list_slug,
        }
        for i in range(1, 12):
            headline = f'Test headline number {str(i)}'
            email = create_email(
                headline=headline,
                list_name=self.subscription.list_name,
                published=True,
            )
            self.email_class = email.__class__
        super().setUp()

    def test_get_invalid_list(self):
        self.kwargs['list_slug'] = 'invalid-slug'
        check_not_found(
            self,
            True,
        )

    def test_get_not_publicly_visible(self):
        self.subscription.publicly_visible = False
        self.subscription.save()
        check_not_found(
            self,
            True,
        )

    def test_get_not_use_pages(self):
        self.subscription.use_pages = False
        self.subscription.save()
        check_not_found(
            self,
            True,
        )

    def test_get_page_content_published(self):
        true_strings = [
            '<a href="/mail_test/subscriptions/subscribe/">',
            '<a href="?q=&amp;page=2">',
        ]
        for i in range(2, 12):
            headline = f'>Test headline number {str(i)}<'
            true_strings.append(headline)
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=true_strings,
            false_strings=[
                '>Test headline number 1<',
            ],
        )

    def test_get_page_content_unpublished(self):
        emails = self.email_class.objects.all()
        for email in emails:
            email.published = False
            email.save()
        true_strings = [
            '<a href="/mail_test/subscriptions/subscribe/">',
        ]
        false_strings = [
            '<a href="?q=&amp;page=',
        ]
        for i in range(1, 12):
            headline = f'>Test headline number {str(i)}<'
            false_strings.append(headline)
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=true_strings,
            false_strings=false_strings,
        )

    def test_get_pagination_page_two(self):
        true_strings = [
            '<a href="/mail_test/subscriptions/subscribe/">',
            '<a href="?q=&amp;page=1">',
            '>Test headline number 1<',
        ]
        false_strings = []
        for i in range(2, 12):
            headline = f'>Test headline number {str(i)}<'
            false_strings.append(headline)
        create_request_response(
            self,
            'get',
            page='2',
        )
        check_http_response(
            self,
            true_strings=true_strings,
            false_strings=false_strings,
        )

    def test_get_pagination_setting(self):
        with self.settings(
            EMAILER_PAGINATION=False,
        ):
            true_strings = [
                '<a href="/mail_test/subscriptions/subscribe/">',
            ]
            for i in range(1, 12):
                headline = f'>Test headline number {str(i)}<'
                true_strings.append(headline)
            create_request_response(
                self,
                'get',
            )
            check_http_response(
                self,
                true_strings=true_strings,
            )

    def test_get_pagination_results_setting(self):
        with self.settings(
            EMAILER_PAGINATION_RESULTS=5,
        ):
            true_strings = [
                '<a href="/mail_test/subscriptions/subscribe/">',
            ]
            for i in range(7, 12):
                headline = f'>Test headline number {str(i)}<'
                true_strings.append(headline)
            false_strings = [
            ]
            for i in range(1, 7):
                headline = f'>Test headline number {str(i)}<'
                false_strings.append(headline)
            create_request_response(
                self,
                'get',
            )
            check_http_response(
                self,
                true_strings=true_strings,
                false_strings=false_strings,
            )

    def test_get_pagination_not_integer(self):
        true_strings = [
            '<a href="/mail_test/subscriptions/subscribe/">',
            '<a href="?q=&amp;page=2">',
        ]
        for i in range(2, 12):
            headline = f'>Test headline number {str(i)}<'
            true_strings.append(headline)
        create_request_response(
            self,
            'get',
            page='invalid',
        )
        check_http_response(
            self,
            true_strings=true_strings,
            false_strings=[
                '>Test headline number 1<',
            ],
        )


class PageViewPreviewBase(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.email = create_email(published=True)
        self.subscription = self.email.subscription_list
        self.subscription_two = create_subscription(list_name='List Two')
        self.kwargs = {
            'list_slug': self.subscription.list_slug,
            'year': '2000',
            'month': '1',
            'day': '1',
            'pk': self.email.pk,
            'headline_slug': 'dummy-headline-slug',
        }
        super().setUp()


class PageViewTests(PageViewPreviewBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.view_name = 'page_view'
        super().setUp()

    def test_get_subscription_invalid_slug(self):
        self.kwargs['list_slug'] = 'invalid-slug'
        check_not_found(
            self,
            True,
        )

    def test_get_subscription_invalid_list(self):
        self.kwargs['list_slug'] = self.subscription_two.list_slug
        check_not_found(
            self,
            True,
        )

    def test_get_subscription_not_public(self):
        self.subscription.publicly_visible = False
        self.subscription.save()
        check_not_found(
            self,
            True,
        )

    def test_get_subscription_not_pages(self):
        self.subscription.use_pages = False
        self.subscription.save()
        check_not_found(
            self,
            True,
        )

    def test_get_page_published(self):
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_strings=[
                'Test headline of many characters',
                '<p>Test body text paragraph one.</p><p>Test body text paragraph two.</p>',
                'Test caption',
                'Test description',
                'Test title',
                'Test extra text',
            ],
        )

    def test_get_page_unpublished(self):
        self.email.published = False
        self.email.save()
        check_not_found(
            self,
            True,
        )

    def test_get_subscription_invalid_pk(self):
        self.kwargs['pk'] = 999
        check_not_found(
            self,
            True,
        )

    def test_get_meta_tags(self):
        create_request_response(
            self,
            'get',
        )
        file_name = self.email.email_image().processed_file.name
        check_http_response(
            self,
            true_strings=[
                '<meta property="og:url" content="http://testserver/mail_test/test-list/2000/1/1/1/dummy-headline-slug.html">',
                '<meta property="og:type" content="article">',
                '<meta property="og:description" content="Test body text paragraph one.">',
                '<meta property="og:title" content="Test headline of many characters">',
                f'<meta property="og:image" content="http://127.0.0.1:8000/media/{file_name}">',
                f'<meta property="og:image:url" content="http://127.0.0.1:8000/media/{file_name}">',
                '<meta property="og:image:type" content="image/png">',
                '<meta property="og:image:width" content="1080">',
                '<meta property="og:image:height" content="1080">',
                '<meta property="og:image:alt" content="Test description">',
            ],
        )

    def test_get_default_image_settings(self):
        with self.settings(
            EMAILER_DEFAULT_IMAGE='test-default-image',
            EMAILER_DEFAULT_TYPE='test-default-type',
            EMAILER_DEFAULT_WIDTH='test-default-width',
            EMAILER_DEFAULT_HEIGHT='test-default-height',
            EMAILER_DEFAULT_ALT='test-default-alt',
        ):
            self.email.email_image().delete()
            create_request_response(
                self,
                'get',
            )
            check_http_response(
                self,
                true_strings=[
                    '<meta property="og:image" content="test-default-image">',
                    '<meta property="og:image:type" content="test-default-type">',
                    '<meta property="og:image:width" content="test-default-width">',
                    '<meta property="og:image:height" content="test-default-height">',
                    '<meta property="og:image:alt" content="test-default-alt">',
                ],
            )


class PagePreviewTests(PageViewPreviewBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.view_name = 'page_preview'
        super().setUp()

    def test_get_preview_without_permission(self):
        self.user = create_user()
        check_permission(
            self,
            False,
        )

    def test_get_preview_with_view_permission(self):
        self.user = create_user(
            permission_list=[
                'view_bulkemail',
            ],
        )
        check_permission(
            self,
            True,
        )

    def test_get_preview_unpublished(self):
        self.email.published = False
        self.email.save()
        self.user = create_user(
            permission_list=[
                'view_bulkemail',
            ],
        )
        check_permission(
            self,
            True,
        )


@patch(
    'django.utils.timezone.now',
    fake_now,
)
class OpenedEmailTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_year = fake_now().year,
        self.current_month = fake_now().month,

    def setUp(self):
        self.subscriber = create_subscriber()
        self.tracker = create_tracker()
        self.view_name = 'opened_email'
        self.kwargs = {
            'pk': self.tracker.pk,
            'subscriber_key': self.subscriber.subscriber_key,
        }
        self.image_dict = {
            'width': 1,
            'height': 1,
            'mode': 'RGBA',
            'format': 'PNG',
        }
        self.test_json_dict = {
            self.subscriber.subscriber_key: [
                self.current_year[0],
                self.current_month[0],
            ]
        }
        super().setUp()

    def test_get_invalid_pk(self):
        self.kwargs['pk'] = 999
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            image_dict=self.image_dict,
        )

    def test_get_no_existing_data(self):
        create_request_response(
            self,
            'get',
        )
        check_http_response(
            self,
            true_dict=self.test_json_dict,
            image_dict=self.image_dict,
        )

    def test_get_existing_data(self):
        mock_json_dict = {
            'test_key': 'Test value',
        }
        self.tracker.json_data = mock_json_dict
        self.tracker.save()
        create_request_response(
            self,
            'get',
        )
        merged_json_dict = {**mock_json_dict, **self.test_json_dict}
        check_http_response(
            self,
            true_dict=merged_json_dict,
            image_dict=self.image_dict,
        )


@patch(
    'mailchimp3.MailChimp',
)
class MCSyncTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.view_name = 'mc_sync'
        self.kwargs = {}
        self.original_email = 'original@example.com'
        self.orignal_first_name = 'OriginalFirst'
        self.orignal_last_name = 'OriginalLast'
        self.updated_email = 'updated@example.com'
        self.updated_first_name = 'UpdatedFirst'
        self.updated_last_name = 'UpdatedLast'
        self.test_list_id = 'testlistid'
        self.subscription_one = create_subscription(
            list_name='List One',
            mc_sync=True,
            mc_list=self.test_list_id,
        )
        self.subscription_two = create_subscription(
            list_name='List Two',
        )
        self.data = {
            'data[list_id]': self.test_list_id,
            'data[email]': self.original_email,
            'data[merges][FNAME]': self.orignal_first_name,
            'data[merges][LNAME]': self.orignal_last_name,
        }
        self.subscriber_attributes = {
            'first_name': self.orignal_first_name,
            'last_name': self.orignal_last_name,
            'mc_email': self.original_email,
        }
        self.states_dict = {
            0: 'does not exist',
            1: 'exists with no subscriptions',
            2: 'exists with subscription one',
            3: 'exists with subscription two',
            4: 'exists with subscriptions one and two',
        }
        self.outgoing_json = ''
        super().setUp()

    def create_url(self):
        url_string = f"{self.profile_instance.protocol}{self.profile_instance.domain}{reverse('django_simple_bulk_emailer:mc_sync')}?key={self.subscription_one.secret_key}"
        return url_string

    def mock_all_webhooks(self, *args, **kwargs):
        incoming_dict = {
            'webhooks': [
                {
                    'url': self.create_url(),
                    'id': 'webhook_ID',
                },
            ],
        }
        return incoming_dict

    def mock_update_webhook(self, *args, **kwargs):
        self.outgoing_json = kwargs

    def test_post_correct_key(self, MockMailChimp):
        create_request_response(
            self,
            'post',
            key=self.subscription_one.secret_key,
        )
        check_http_response(
            self,
            true_strings=[
                'COMPLETED',
            ],
        )

    def test_post_invalid_key(self, MockMailChimp):
        create_request_response(
            self,
            'post',
            key='incorrect_key',
        )
        check_http_response(
            self,
            true_strings=[
                'INVALID CREDENTIALS',
            ],
        )

    def test_post_no_key(self, MockMailChimp):
        create_request_response(
            self,
            'post',
        )
        check_http_response(
            self,
            true_strings=[
                'INVALID CREDENTIALS',
            ],
        )

    def test_post_subscribe(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data['type'] = 'subscribe'
                state_comparisons = {
                    0: 2,
                    1: 2,
                    2: 2,
                    3: 4,
                    4: 4,
                }
                for start_state in state_comparisons.keys():
                    create_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.orignal_first_name,
                        self.orignal_last_name,
                        start_state,
                    )
                    self.subscription_one.refresh_from_db()
                    create_request_response(
                        self,
                        'post',
                        key=self.subscription_one.secret_key,
                    )
                    extra_text = f" — state tested was '{self.states_dict.get(start_state)}'"
                    check_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.subscriber_attributes,
                        state_comparisons.get(
                            start_state,
                        ),
                        extra_text=extra_text,
                    )
                    remove_subscriber(self.original_email)

    def test_post_unsubscribe(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data['type'] = 'unsubscribe'
                state_comparisons = {
                    0: 0,
                    1: 1,
                    2: 1,
                    3: 3,
                    4: 3,
                }
                for start_state in state_comparisons.keys():
                    create_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.orignal_first_name,
                        self.orignal_last_name,
                        start_state,
                    )
                    self.subscription_one.refresh_from_db()
                    create_request_response(
                        self,
                        'post',
                        key=self.subscription_one.secret_key,
                    )
                    extra_text = f" — state tested was '{self.states_dict.get(start_state)}'"
                    check_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.subscriber_attributes,
                        state_comparisons.get(
                            start_state,
                        ),
                        extra_text=extra_text,
                    )
                    remove_subscriber(self.original_email)

    def test_post_cleaned(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data = {
                    'type': 'cleaned',
                    'data[list_id]': self.test_list_id,
                    'data[email]': self.original_email,
                }
                state_comparisons = {
                    0: 0,
                    1: 1,
                    2: 1,
                    3: 3,
                    4: 3,
                }
                for start_state in state_comparisons.keys():
                    create_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.orignal_first_name,
                        self.orignal_last_name,
                        start_state,
                    )
                    self.subscription_one.refresh_from_db()
                    create_request_response(
                        self,
                        'post',
                        key=self.subscription_one.secret_key,
                    )
                    extra_text = f" — state tested was '{self.states_dict.get(start_state)}'"
                    check_subscriber_subscription_state(
                        self,
                        self.original_email,
                        self.subscriber_attributes,
                        state_comparisons.get(
                            start_state,
                        ),
                        extra_text=extra_text,
                    )
                    remove_subscriber(self.original_email)

    def test_post_update_email(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data = {
                    'type': 'upemail',
                    'data[list_id]': self.test_list_id,
                    'data[new_email]': self.updated_email,
                    'data[old_email]': self.original_email,
                }
                updated_subscriber_attributes = {
                    'mc_email': self.updated_email,
                }
                original_state_comparisons = {
                    0: 0,
                    1: 0,
                    2: 0,
                    3: 0,
                    4: 0,
                }
                combined_state_comparisons = {
                    0: {
                        0: 2,
                        1: 2,
                        2: 2,
                        3: 4,
                        4: 4,
                    },
                    1: {
                        0: 2,
                        1: 2,
                        2: 2,
                        3: 4,
                        4: 4,
                    },
                    2: {
                        0: 2,
                        1: 2,
                        2: 2,
                        3: 4,
                        4: 4,
                    },
                    3: {
                        0: 4,
                        1: 4,
                        2: 4,
                        3: 4,
                        4: 4,
                    },
                    4: {
                        0: 4,
                        1: 4,
                        2: 4,
                        3: 4,
                        4: 4,
                    },
                }
                for original_start_state in original_state_comparisons.keys():
                    updated_state_comparisons = combined_state_comparisons.get(
                        original_start_state,
                    )
                    for updated_start_state in updated_state_comparisons.keys():
                        create_subscriber_subscription_state(
                            self,
                            self.original_email,
                            self.orignal_first_name,
                            self.orignal_last_name,
                            original_start_state,
                        )
                        create_subscriber_subscription_state(
                            self,
                            self.updated_email,
                            self.orignal_first_name,
                            self.orignal_last_name,
                            updated_start_state,
                        )
                        self.subscription_one.refresh_from_db()
                        create_request_response(
                            self,
                            'post',
                            key=self.subscription_one.secret_key,
                        )
                        extra_text = f" — old state was '{self.states_dict.get(original_start_state)}' and new state was '{self.states_dict.get(updated_start_state)}'"
                        check_subscriber_subscription_state(
                            self,
                            self.original_email,
                            self.subscriber_attributes,
                            original_state_comparisons.get(
                                original_start_state,
                            ),
                            extra_text=extra_text,
                        )
                        check_subscriber_subscription_state(
                            self,
                            self.updated_email,
                            updated_subscriber_attributes,
                            updated_state_comparisons.get(
                                updated_start_state,
                            ),
                            extra_text=extra_text,
                        )
                        remove_subscriber(self.original_email)
                        remove_subscriber(self.updated_email)

    def test_post_update_profile_existing(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data = {
                    'type': 'profile',
                    'data[list_id]': self.test_list_id,
                    'data[email]': self.original_email,
                    'data[merges][FNAME]': self.updated_first_name,
                    'data[merges][LNAME]': self.updated_last_name,
                }
                updated_subscriber_attributes = {
                    'first_name': self.updated_first_name,
                    'last_name': self.updated_last_name,
                    'subscription_choices': [str(self.subscription_one.pk)],
                }
                create_subscriber(
                    subscriber_email=self.original_email,
                    first_name=self.orignal_first_name,
                    last_name=self.orignal_last_name,
                )
                create_request_response(
                    self,
                    'post',
                    key=self.subscription_one.secret_key,
                )
                check_subscriber_attributes(
                    self,
                    self.original_email,
                    updated_subscriber_attributes,
                    True,
                )

    def test_post_update_profile_not_existing(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data = {
                    'type': 'profile',
                    'data[list_id]': self.test_list_id,
                    'data[email]': self.original_email,
                    'data[merges][FNAME]': self.updated_first_name,
                    'data[merges][LNAME]': self.updated_last_name,
                }
                updated_subscriber_attributes = {
                    'first_name': self.updated_first_name,
                    'last_name': self.updated_last_name,
                    'subscription_choices': [str(self.subscription_one.pk)],
                }
                create_request_response(
                    self,
                    'post',
                    key=self.subscription_one.secret_key,
                )
                check_subscriber_attributes(
                    self,
                    self.original_email,
                    updated_subscriber_attributes,
                    True,
                )

    def test_local_key_updated(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                old_key = self.subscription_one.secret_key
                self.data = {
                    'type': 'profile',
                    'data[list_id]': self.test_list_id,
                    'data[email]': self.original_email,
                    'data[merges][FNAME]': self.updated_first_name,
                    'data[merges][LNAME]': self.updated_last_name,
                }
                create_request_response(
                    self,
                    'post',
                    key=self.subscription_one.secret_key,
                )
                self.subscription_one.refresh_from_db()
                new_key = self.subscription_one.secret_key
                compare_secret_keys(
                    self,
                    old_key,
                    new_key,
                )

    def test_outgoing_json(self, MockMailChimp):
        with patch.object(MockMailChimp().lists.webhooks, 'all', new=self.mock_all_webhooks):
            with patch.object(MockMailChimp().lists.webhooks, 'update', new=self.mock_update_webhook):
                self.data = {
                    'type': 'profile',
                    'data[list_id]': self.test_list_id,
                    'data[email]': self.original_email,
                    'data[merges][FNAME]': self.updated_first_name,
                    'data[merges][LNAME]': self.updated_last_name,
                }
                create_request_response(
                    self,
                    'post',
                    key=self.subscription_one.secret_key,
                )
                self.subscription_one.refresh_from_db()
                mock_data = {
                    'list_id': 'testlistid',
                    'webhook_id': 'webhook_ID',
                    'data': {
                        'url': f'{self.create_url()}'
                    },
                }
                json_contains(
                    self,
                    json_data=self.outgoing_json,
                    true_dict=mock_data,
                )



from django.core.management.base import (
    BaseCommand,
)
from django.utils import (
    timezone,
)


from ...models import (
    BulkEmail,
)


class Command(BaseCommand):
    help = 'Deletes emails that have reached or passed their deletion date'

    def handle(self, *args, **options):
        BulkEmail.objects.filter(
            deletion_date__lte=timezone.now().date(),
        ).delete()



import hashlib


from django.core.management.base import (
    BaseCommand,
)


from mailchimp3 import (
    MailChimp,
)
from mailchimp3.mailchimpclient import (
    MailChimpError,
)


from ...models import (
    Subscriber,
    Subscription,
)


def get_client(subscription):
    client = MailChimp(
        mc_api=subscription.mc_api,
        mc_user=subscription.mc_user,
        timeout=30.0,
    )
    return client


def get_hash(email_address):
    return hashlib.md5(email_address.encode('utf-8')).hexdigest()


class Command(BaseCommand):
    help = 'Syncs subscriber data to MailChimp'

    def handle(self, *args, **options):
        """ """
        ''' Get subscribers that are not yet synced '''
        subscribers = Subscriber.objects.filter(
            mc_synced=False,
        )
        ''' Get subscriptions that are set up for syncing '''
        all_subscriptions = Subscription.objects.filter(
            mc_sync=True,
        )
        ''' For each subscriber, try to create or update a subscribed MailChimp contact '''
        for subscriber in subscribers:
            ''' Try to pass subscriber data to MailChimp client using mc_email address hash '''
            subscriber_hash = get_hash(subscriber.mc_email)
            for subscription in all_subscriptions:
                client = get_client(subscription)
                subscriber_data = {
                    'status_if_new': 'subscribed',
                    'status': 'subscribed',
                    'email_address': subscriber.subscriber_email,
                    'merge_fields': {
                        'FNAME': subscriber.first_name,
                        'LNAME': subscriber.last_name,
                    },
                }
                try:
                    client.lists.members.create_or_update(
                        list_id=subscription.mc_list,
                        subscriber_hash=subscriber_hash,
                        data=subscriber_data,
                    )
                    print(f'Created or updated contact {subscriber.mc_email} as {subscriber.subscriber_email} in list {subscription.list_name}')
                except MailChimpError:
                    print(f'Unable to create or update contact {subscriber.mc_email} as {subscriber.subscriber_email} in list {subscription.list_name}')
                    try:
                        ''' If unable to pass subscriber data to MailChimp client, try subscriber_email address hash '''
                        subscriber_hash = get_hash(subscriber.subscriber_email)
                        client.lists.members.create_or_update(
                            list_id=subscription.mc_list,
                            subscriber_hash=subscriber_hash,
                            data=subscriber_data,
                        )
                        print(f'Created or updated contact {subscriber.subscriber_email} in list {subscription.list_name}')
                    except MailChimpError:
                        print(f'Unable to create or update contact {subscriber.subscriber_email} in list {subscription.list_name}')
                        ''' Remove subscription from subscriber if cannot be synced to MailChimp '''
                        subscriber.subscriptions.remove(subscription)
                        print(f'Removed subscription {subscription.list_name} from subscriber {subscriber.subscriber_email}')

            ''' For each nonsubscribed subscriber, try to unsubscribe its corresponding MailChimp contact '''
            subscriber_hash = get_hash(subscriber.subscriber_email)
            subscriber_subscriptions = subscriber.subscriptions.filter(
                mc_sync=True,
            )
            for subscription in all_subscriptions:
                if subscription not in subscriber_subscriptions:
                    client = get_client(subscription)
                    subscriber_data = {
                        'status': 'unsubscribed',
                    }
                    try:
                        client.lists.members.update(
                            list_id=subscription.mc_list,
                            subscriber_hash=subscriber_hash,
                            data=subscriber_data,
                        )
                        print(f'Unsubscribed contact {subscriber.subscriber_email} from list {subscription.list_name}')
                    except MailChimpError:
                        print(f'Unable to unsubscribe contact {subscriber.subscriber_email} from list {subscription.list_name}')
            ''' Update the subscriber object's attributes '''
            subscriber.mc_email = subscriber.subscriber_email
            subscriber.mc_synced = True
            subscriber.save()
            print(f'Completed syncing MailChimp for contact {subscriber.subscriber_email} in list {subscription.list_name}')



from io import (
    BytesIO,
)


from django.conf import (
    settings,
)
from django.contrib.sites.models import (
    Site,
)
from django.core.exceptions import (
    ObjectDoesNotExist,
    PermissionDenied,
)
from django.core.mail import (
    EmailMultiAlternatives,
)
from django.core.paginator import (
    EmptyPage,
    PageNotAnInteger,
    Paginator,
)
from django.db import (
    IntegrityError,
)
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseRedirect,
)
from django.shortcuts import (
    render,
)
from django.template.loader import (
    get_template,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone,
)
from django.views.decorators.csrf import (
    csrf_exempt,
)


from PIL import (
    Image,
)


from .forms import (
    GetSubscriberForm,
    ModifySubscriberForm,
)
from .models import (
    EmailTracker,
    SiteProfile,
    Subscriber,
    Subscription,
)


def get_universal_email_directory():
    try:
        return settings.EMAILER_EMAIL_TEMPLATES
    except AttributeError:
        return 'django_simple_bulk_emailer/universal/emails'


def get_universal_page_directory():
    try:
        return settings.EMAILER_PAGE_TEMPLATES
    except AttributeError:
        return 'django_simple_bulk_emailer/universal/pages'


def send_email(email_content, list_slug='', subscriber_key='', subject='', text_template='', html_template='', to_address=''):
    try:
        from_address = settings.EMAILER_FROM_ADDRESS
    except AttributeError:
        from_address = settings.DEFAULT_FROM_EMAIL
    try:
        reply_address = settings.EMAILER_REPLY_ADDRESS
    except AttributeError:
        reply_address = from_address
    site_domain = Site.objects.get_current().domain
    site_profile = SiteProfile.objects.filter(
        domain=site_domain,
    ).first()
    protocol_domain = site_profile.protocol_domain()
    manage_url = reverse(
        'django_simple_bulk_emailer:manage_subscriptions',
        kwargs={
            'subscriber_key': subscriber_key,
        },
    )
    if list_slug != '':
        unsubscribe_url = reverse(
            'django_simple_bulk_emailer:quick_unsubscribe',
            kwargs={
                'list_slug': list_slug,
                'subscriber_key': subscriber_key,
            },
        )
    else:
        unsubscribe_url = ''
    subscriptions_url = f'{protocol_domain}{manage_url}'
    quick_unsubscribe_url = f'{protocol_domain}{unsubscribe_url}'
    email_content['subscriptions_url'] = subscriptions_url
    email_content['quick_unsubscribe_url'] = quick_unsubscribe_url
    text_email = get_template(text_template).render(email_content)
    html_email = get_template(html_template).render(email_content)
    message = EmailMultiAlternatives(
        subject,
        text_email,
        from_address,
        [to_address],
        reply_to=[reply_address],
    )
    message.attach_alternative(html_email, 'text/html')
    try:
        message.send(
            fail_silently=True,
        )
    except IOError:
        pass


def get_subscriptions(request):
    """ """
    ''' Begin getting time elements '''
    form_submit_time = timezone.now().timestamp()
    if 'form_load_time' in request.session:
        form_load_time = request.session['form_load_time']
        del request.session['form_load_time']
    else:
        form_load_time = form_submit_time
    ''' End getting time elements '''
    form = GetSubscriberForm
    if request.method == 'POST':
        form = GetSubscriberForm(request.POST)
        if form.is_valid():
            success_template = f'{get_universal_page_directory()}/subscribe_success.html'
            ''' Begin checking for input in hidden field '''
            email = form.cleaned_data['email']
            if email:
                return render(
                    request,
                    success_template,
                )
            ''' End checking for input in hidden field '''
            ''' Begin checking how quickly form was submitted '''
            if form_submit_time - form_load_time < 1:
                return render(
                    request,
                    success_template,
                )
            ''' End checking how quickly form was submitted '''
            subscriber_email = form.cleaned_data['subscriber_email'].lower()
            email_content = {}
            subscriber, created = Subscriber.objects.get_or_create(
                subscriber_email=subscriber_email,
            )
            text_template = f'{get_universal_email_directory()}/subscribe_email_text.txt'
            html_template = f'{get_universal_email_directory()}/subscribe_email_html.html'
            try:
                subject = settings.EMAILER_SUBSCRIBE_SUBJECT
            except AttributeError:
                subject = 'Manage your email subscriptions'
            send_email(
                email_content,
                subscriber_key=subscriber.subscriber_key,
                text_template=text_template,
                html_template=html_template,
                subject=subject,
                to_address=subscriber_email,
            )
            return render(
                request,
                success_template,
            )
    ''' Begin setting when form was submitted '''
    request.session['form_load_time'] = timezone.now().timestamp()
    ''' End setting when form was submitted '''
    page_template = f'{get_universal_page_directory()}/subscribe.html'
    subscription_set = Subscription.objects.filter(
        publicly_visible=True,
    )
    context = {
        'form': form,
        'subscription_set': subscription_set,
    }
    return render(
        request,
        page_template,
        context,
    )


def manage_subscriptions(request, subscriber_key):
    try:
        subscriber = Subscriber.objects.get(
            subscriber_key=subscriber_key,
        )
        subscription_set = subscriber.subscriptions.all()
        form = ModifySubscriberForm(
            initial={
                'subscription_choices': subscription_set,
            },
            instance=subscriber,
        )
        if request.method == 'POST':
            success_template = f'{get_universal_page_directory()}/manage_subscriptions_success.html'
            if 'unsubscribe_all' in request.POST:
                subscriber.subscriptions.clear()
                return render(
                    request,
                    success_template,
                )
            form = ModifySubscriberForm(
                request.POST,
                instance=subscriber,
            )
            if form.is_valid():
                form.save(commit=False)
                subscriber.subscriptions.set(form.cleaned_data['subscription_choices'])
                subscriber.save()
                return render(
                    request,
                    success_template,
                )
        page_template = f'{get_universal_page_directory()}/manage_subscriptions.html'
        context = {
            'form': form,
            'subscriber': subscriber,
            'subscription_set': subscription_set,
        }
        return render(
            request,
            page_template,
            context,
        )
    except ObjectDoesNotExist:
        subscribe_url = reverse(
            'django_simple_bulk_emailer:get_subscriptions',
        )
        page_template = f'{get_universal_page_directory()}/manage_subscriptions_error.html'
        context = {
            'subscribe_url': subscribe_url,
        }
        return render(
            request,
            page_template,
            context,
        )


def quick_unsubscribe(request, list_slug, subscriber_key):
    try:
        subscriber = Subscriber.objects.get(
            subscriber_key=subscriber_key,
        )
        subscription = subscriber.subscriptions.get(
            list_slug=list_slug,
        )
        subscriber.subscriptions.remove(subscription)
        manage_url = reverse(
            'django_simple_bulk_emailer:manage_subscriptions',
            kwargs={
                'subscriber_key': subscriber.subscriber_key,
            }
        )
        page_template = f'{get_universal_page_directory()}/quick_unsubscribe.html'
        context = {
            'list_name': subscription.list_name,
            'manage_url': manage_url,
        }
        return render(
            request,
            page_template,
            context,
        )
    except ObjectDoesNotExist:
        subscribe_url = reverse(
            'django_simple_bulk_emailer:get_subscriptions',
        )
        page_template = f'{get_universal_page_directory()}/manage_subscriptions_error.html'
        context = {
            'subscribe_url': subscribe_url,
        }
        return render(
            request,
            page_template,
            context,
        )


def email_preview(request, list_slug, pk):
    if not request.user.has_perm('django_simple_bulk_emailer.change_bulkemail') \
            and not request.user.has_perm('django_simple_bulk_emailer.view_bulkemail'):
        raise PermissionDenied
    subscription = Subscription.objects.filter(
        list_slug=list_slug,
    ).first()
    if not subscription:
        raise Http404()
    else:
        email_class = subscription.get_email_class()
        try:
            email_instance = email_class.objects.get(
                pk=pk,
            )
        except (ObjectDoesNotExist, ValueError) as e:
            raise Http404()
        list_return = HttpResponseRedirect(
            reverse(
                f'admin:{email_instance._meta.app_label}_{email_instance._meta.model_name}_changelist'
            )
        )
        page_template = f'{subscription.email_directory}/email_template_html.html'
        basic_template = f'{get_universal_page_directory()}/bulk_email_preview.html'
        if request.method == "POST":
            if 'send_email' in request.POST:
                email_instance.sendable = True
                email_instance.sending = True
                email_instance.sent = True
                email_instance.save()
                return list_return
            if 'return_list' in request.POST:
                return list_return
        context = {
            'basic_template': basic_template,
            'email_instance': email_instance,
        }
        return render(
            request,
            page_template,
            context,
        )


def list_view(request, list_slug):
    subscription = Subscription.objects.filter(
        list_slug=list_slug,
        publicly_visible=True,
        use_pages=True,
    ).first()
    if not subscription:
        raise Http404()
    else:
        email_class = subscription.get_email_class()
        email_set = email_class.objects.filter(
            subscription_list=subscription,
            published=True,
        ).order_by(
            '-publication_date',
            '-created',
        )
        try:
            paginate = settings.EMAILER_PAGINATION
        except AttributeError:
            paginate = True
        if paginate:
            try:
                page_results = settings.EMAILER_PAGINATION_RESULTS
            except AttributeError:
                page_results = 10
            paginator = Paginator(email_set, page_results)
            page = request.GET.get('page')
            try:
                email_set = paginator.page(page)
            except PageNotAnInteger:
                email_set = paginator.page(1)
            except EmptyPage:
                email_set = paginator.page(paginator.num_pages)
        page_template = f'{subscription.page_directory}/list_view.html'
        subscribe_url = reverse(
            'django_simple_bulk_emailer:get_subscriptions',
        )
        context = {
            'subscription': subscription,
            'email_set': email_set,
            'subscribe_url': subscribe_url,
        }
        return render(
            request,
            page_template,
            context,
        )


def page_view(request, list_slug, year, month, day, pk, headline_slug, preview=False):
    subscription = Subscription.objects.filter(
        list_slug=list_slug,
        publicly_visible=True,
        use_pages=True,
    ).first()
    if not subscription:
        raise Http404()
    else:
        email_class = subscription.get_email_class()
        try:
            email_instance = email_class.objects.get(
                subscription_list=subscription,
                pk=pk,
            )
        except (ObjectDoesNotExist, ValueError) as e:
            raise Http404()
        if not preview and not email_instance.published:
            raise Http404()
        page_template = f'{subscription.page_directory}/page_view.html'
        context = {
            'email_instance': email_instance,
        }
        context_defaults_dict = {
            'default_image': 'EMAILER_DEFAULT_IMAGE',
            'default_type': 'EMAILER_DEFAULT_TYPE',
            'default_width': 'EMAILER_DEFAULT_WIDTH',
            'default_height': 'EMAILER_DEFAULT_HEIGHT',
            'default_alt': 'EMAILER_DEFAULT_ALT',
        }
        for key, value in context_defaults_dict.items():
            try:
                new_value = getattr(settings, value)
            except AttributeError:
                new_value = ''
            context.update({key: new_value})
        return render(
            request,
            page_template,
            context,
        )


def page_preview(request, list_slug, year, month, day, pk, headline_slug):
    if not request.user.has_perm('django_simple_bulk_emailer.change_bulkemail') \
            and not request.user.has_perm('django_simple_bulk_emailer.view_bulkemail'):
        raise PermissionDenied
    return page_view(request, list_slug, year, month, day, pk, headline_slug, preview=True)


def opened_email(request, pk, subscriber_key):
    try:
        email_tracker = EmailTracker.objects.get(
            pk=pk,
        )
        if email_tracker.json_data:
            dict_data = email_tracker.json_data
        else:
            dict_data = {}
        if subscriber_key not in dict_data:
            dict_data[subscriber_key] = [
                timezone.now().year,
                timezone.now().month,
            ]
            email_tracker.json_data = dict_data
            email_tracker.save()
    except (ObjectDoesNotExist, ValueError) as e:
        pass
    temp_handle = BytesIO()
    image_file = Image.new(
        'RGBA',
        (1, 1),
        (0, 0, 0, 0),
    )
    image_file.save(
        temp_handle,
        'png',
    )
    temp_handle.seek(0)
    image_data = temp_handle.read()
    return HttpResponse(
        image_data,
        content_type='image/png',
    )


def get_subscriber(email):
    return Subscriber.objects.get(
        subscriber_email=email,
    )


def new_subscriber(email):
    return Subscriber(
        subscriber_email=email,
        )


@csrf_exempt
def mc_sync(request):
    try:
        request_type = request.POST.get(
            'type',
            '',
        )
        email = request.POST.get(
            'data[email]',
            '',
        ).lower()
        old_email = request.POST.get(
            'data[old_email]',
            '',
        ).lower()
        new_email = request.POST.get(
            'data[new_email]',
            '',
        ).lower()
        first_name = request.POST.get(
            'data[merges][FNAME]',
            '',
        )
        last_name = request.POST.get(
            'data[merges][LNAME]',
            '',
        )
        mc_list = request.POST.get(
            'data[list_id]',
            '',
        )
        secret_key = request.GET.get('key', '')
        subscription = Subscription.objects.filter(
            mc_sync=True,
        ).filter(
            mc_list=mc_list,
        ).filter(
            secret_key=secret_key,
        ).first()
        if subscription:
            if request_type == 'upemail':
                try:
                    subscriber_old = get_subscriber(old_email)
                except ObjectDoesNotExist:
                    subscriber_old = None
                try:
                    subscriber_new = get_subscriber(new_email)
                except ObjectDoesNotExist:
                    subscriber_new = None
                if subscriber_old is not None:
                    if subscriber_new:
                        for subscription_item in subscriber_new.subscriptions.all():
                            subscriber_old.subscriptions.add(subscription_item)
                            subscriber_new.subscriptions.remove(subscription_item)
                        subscriber_new.delete()
                    subscriber_old.subscriber_email = new_email
                    subscriber_old.mc_email = new_email
                    subscriber_old.save()
                    subscriber_old.subscriptions.add(subscription)
                else:
                    if subscriber_new is None:
                        subscriber_new = new_subscriber(new_email)
                        subscriber_new.save()
                    subscriber_new.subscriptions.add(subscription)
            else:
                if request_type == 'unsubscribe' or request_type == 'cleaned':
                    try:
                        subscriber = get_subscriber(email)
                        subscriber.subscriptions.remove(subscription)
                    except ObjectDoesNotExist:
                        pass
                else:
                    try:
                        subscriber = get_subscriber(email)
                    except ObjectDoesNotExist:
                        subscriber = new_subscriber(email)
                    if first_name != '':
                        subscriber.first_name = first_name
                    if last_name != '':
                        subscriber.last_name = last_name
                    subscriber.save()
                    subscriber.subscriptions.add(subscription)
            secret_key_old = subscription.secret_key
            subscription.secret_key = ''
            subscription.save()
            site = Site.objects.get(
                id=settings.SITE_ID,
            )
            site_profile = SiteProfile.objects.filter(
                domain=site.domain,
            ).first()
            base_url = reverse(
                'django_simple_bulk_emailer:mc_sync',
            )
            url_old = f'{site_profile.protocol_domain()}{base_url}?key={secret_key_old}'
            url_new = f'{site_profile.protocol_domain()}{base_url}?key={subscription.secret_key}'
            try:
                from mailchimp3 import (
                    MailChimp,
                )
                from mailchimp3.mailchimpclient import (
                    MailChimpError,
                )
                client = MailChimp(
                    mc_api=subscription.mc_api,
                    mc_user=subscription.mc_user,
                )
                try:
                    webhooks = client.lists.webhooks.all(list_id=mc_list)['webhooks']
                    for hook in webhooks:
                        if hook['url'] == url_old:
                            hook_id = hook['id']
                            client.lists.webhooks.update(
                                list_id=mc_list,
                                webhook_id=hook_id,
                                data={
                                    'url': url_new,
                                }
                            )
                            break
                except MailChimpError:
                    pass
            except ImportError:
                pass
        else:
            return HttpResponse('INVALID CREDENTIALS')
    except IntegrityError:
        pass
    return HttpResponse('COMPLETED')



from django.urls import (
    include,
    path,
)
from django.contrib import (
    admin,
)


urlpatterns = [
      path(
          'admin/',
          admin.site.urls,
      ),
      path(
          'mail_test/',
          include('django_simple_bulk_emailer.urls'),
      ),
]



import ckeditor.fields
import django.contrib.sites.models
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_simple_bulk_emailer.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('django_simple_file_handler', '0006_auto_20190429_1949'),
    ]

    operations = [
        migrations.CreateModel(
            name='BulkEmail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('headline', models.CharField(max_length=254)),
                ('body_text', ckeditor.fields.RichTextField()),
                ('publication_date', models.DateField(default=django.utils.timezone.localdate)),
                ('deletion_date', models.DateField(blank=True, default=django_simple_bulk_emailer.models.get_deletion_date, null=True)),
                ('published', models.BooleanField(default=False)),
                ('is_updated', models.BooleanField(default=False, verbose_name='has been updated')),
                ('sendable', models.BooleanField(default=False)),
                ('sending', models.BooleanField(default=False)),
                ('sent', models.BooleanField(default=False)),
                ('send_history', models.TextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EmailTracker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('subject', models.CharField(max_length=254)),
                ('subscription_name', models.CharField(max_length=254)),
                ('send_complete', models.DateTimeField(default=django.utils.timezone.now)),
                ('number_sent', models.PositiveIntegerField(default=0)),
                ('json_data', models.TextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SiteProfile',
            fields=[
                ('site_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='sites.Site')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('protocol', models.CharField(default='https://', max_length=254)),
            ],
            options={
                'abstract': False,
            },
            bases=('sites.site', models.Model),
            managers=[
                ('objects', django.contrib.sites.models.SiteManager()),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('list_name', models.CharField(max_length=254)),
                ('descriptive_text', ckeditor.fields.RichTextField(blank=True, verbose_name='descriptive text (if using page views)')),
                ('list_slug', models.CharField(blank=True, max_length=254)),
                ('publicly_visible', models.BooleanField(default=False)),
                ('use_pages', models.BooleanField(default=True, verbose_name='use page view')),
                ('email_directory', models.CharField(default='django_simple_bulk_emailer/subscription/emails', max_length=254, verbose_name='email template directory')),
                ('page_directory', models.CharField(default='django_simple_bulk_emailer/subscription/pages', max_length=254, verbose_name='page template directory')),
                ('associated_model', models.CharField(default='django_simple_bulk_emailer.models.BulkEmail', max_length=254)),
                ('mc_sync', models.BooleanField(default=False, verbose_name='MailChimp sync')),
                ('mc_user', models.CharField(default='username', max_length=254, verbose_name='MailChimp username')),
                ('mc_api', models.CharField(default='API_key', max_length=254, verbose_name='MailChimp API key')),
                ('mc_list', models.CharField(default='list_ID', max_length=254, verbose_name='MailChimp audience ID')),
                ('sort_order', models.PositiveIntegerField(default=0, verbose_name='order')),
            ],
            options={
                'ordering': ['sort_order'],
            },
        ),
        migrations.CreateModel(
            name='Subscriber',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('subscriber_key', models.CharField(max_length=254)),
                ('first_name', models.CharField(default='Anonymous', max_length=254)),
                ('last_name', models.CharField(default='Subscriber', max_length=254)),
                ('subscriber_email', models.EmailField(max_length=254, unique=True, verbose_name='email address')),
                ('mc_email', models.EmailField(blank=True, max_length=254, verbose_name='MailChimp email address')),
                ('mc_synced', models.BooleanField(default=False)),
                ('subscriptions', models.ManyToManyField(blank=True, to='django_simple_bulk_emailer.Subscription')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='MonthlyStat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('year_int', models.PositiveIntegerField()),
                ('month_int', models.PositiveIntegerField()),
                ('stat_data', models.TextField(blank=True)),
                ('current_trackers', models.ManyToManyField(related_name='current', to='django_simple_bulk_emailer.EmailTracker')),
                ('older_trackers', models.ManyToManyField(related_name='older', to='django_simple_bulk_emailer.EmailTracker')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EmailImage',
            fields=[
                ('processedimage_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_simple_file_handler.ProcessedImage')),
                ('description', models.CharField(default='Image', max_length=254, verbose_name='screen reader description')),
                ('caption', models.TextField(blank=True, verbose_name='image caption (optional)')),
                ('image_width', models.PositiveIntegerField()),
                ('bulk_email', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='emailimage', to='django_simple_bulk_emailer.BulkEmail')),
            ],
            options={
                'verbose_name': 'image (optional)',
                'verbose_name_plural': 'image (optional)',
            },
            bases=('django_simple_file_handler.processedimage',),
        ),
        migrations.CreateModel(
            name='EmailDocument',
            fields=[
                ('temporarydocument_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_simple_file_handler.TemporaryDocument')),
                ('sort_order', models.PositiveIntegerField(default=0, verbose_name='order')),
                ('bulk_email', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='documents', to='django_simple_bulk_emailer.BulkEmail')),
            ],
            options={
                'verbose_name': 'document',
                'ordering': ['sort_order'],
            },
            bases=('django_simple_file_handler.temporarydocument',),
        ),
        migrations.AddField(
            model_name='bulkemail',
            name='subscription_list',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_simple_bulk_emailer.Subscription'),
        ),
    ]



from django.conf import (
    settings,
)
from django.core.management.base import (
    BaseCommand,
)
from django.utils import (
    timezone,
)


from ...models import (
    EmailTracker,
    MonthlyStat,
)


class Command(BaseCommand):
    help = 'Updates monthly stats with data from email trackers and deletes outdated trackers'

    def handle(self, *args, **options):
        current_datetime = timezone.now()
        monthly_stat, created = MonthlyStat.objects.get_or_create(
            year_int=current_datetime.year,
            month_int=current_datetime.month,
        )
        try:
            tracking_months = settings.EMAILER_TRACKING_MONTHS
        except AttributeError:
            tracking_months = 3
        for tracker in EmailTracker.objects.all():
            ''' Find and delete outdated trackers '''
            deletion_datetime_months = tracker.send_complete.year * 12 + tracker.send_complete.month + tracking_months
            current_datetime_months = current_datetime.year * 12 + current_datetime.month + 1
            if current_datetime_months > deletion_datetime_months:
                tracker.delete()
            else:
                ''' Find and attach any appropriate unattached trackers as current or older '''
                if tracker not in monthly_stat.current_trackers.all() \
                        and tracker not in monthly_stat.older_trackers.all() \
                        and tracker.json_data:
                    for key, value in tracker.json_data.items():
                        year = value[0]
                        month = value[1]
                        if year == current_datetime.year and month == current_datetime.month:
                            if current_datetime.year == tracker.send_complete.year \
                                    and current_datetime.month == tracker.send_complete.month:
                                monthly_stat.current_trackers.add(tracker)
                            else:
                                monthly_stat.older_trackers.add(tracker)
                            break
        ''' Create sorted list of subscription names '''
        subscription_names = []
        month_trackers = monthly_stat.current_trackers.all()
        for tracker in month_trackers:
            if tracker.subscription_name not in subscription_names:
                subscription_names.append(tracker.subscription_name)
        subscription_names.sort()
        ''' Append 'Older emails' to list of subscriptions if appropriate '''
        if monthly_stat.older_trackers.exists():
            subscription_names += ['Older emails']
        ''' Reset stat data and tallies '''
        monthly_stat.stat_data = ''
        for name in subscription_names:
            sent_tally = 0
            open_tally = 0
            monthly_stat.stat_data = \
                f'{monthly_stat.stat_data}' \
                f'<tr id="emailer_title_row">' \
                f'<td>&nbsp;</td>' \
                f'<td>{name}</td>' \
                f'<td id="emailer_numerical">Sent</td>' \
                f'<td id="emailer_numerical">Opens</td>' \
                f'<td></td>' \
                f'</tr>'
            ''' Get appropriate group of trackers '''
            if name == 'Older emails':
                subscription_trackers = monthly_stat.older_trackers.all()
            else:
                subscription_trackers = month_trackers.filter(
                    subscription_name=name,
                )
            stat_dict = {}
            for tracker in subscription_trackers:
                ''' Calculate tracker's opened number for current month '''
                if tracker.json_data:
                    opens = 0
                    for key, value in tracker.json_data.items():
                        year = value[0]
                        month = value[1]
                        if year == current_datetime.year and month == current_datetime.month:
                            opens += 1
                        ''' Create a list of stat data and add to stat dictionary '''
                        stat_dict[tracker.pk] = [
                            opens,
                            tracker.number_sent,
                            tracker.subject,
                            tracker.send_complete_string(),
                        ]
            ''' Sort stat dictionary '''
            sorted_by_value = sorted(stat_dict.items(), key=lambda kv: kv[1], reverse=True)
            if sorted_by_value:
                ''' Format data as rows '''
                row_number = 1
                for email in sorted_by_value:
                    if row_number & 1:
                        row_id = 'emailer_row_odd'
                    else:
                        row_id = 'emailer_row_even'
                    row_number_str = f'{str(row_number)}.'
                    tracker_data = email[1]
                    opens = f'{tracker_data[0]:,}'
                    sent = f'{tracker_data[1]:,}'
                    if tracker_data[0] != 0 and tracker_data[1] != 0:
                        percentage = f'{tracker_data[0] / tracker_data[1]:.1%}'
                    else:
                        percentage = '0%'
                    subject = tracker_data[2]
                    distribution_date = tracker_data[3]
                    monthly_stat.stat_data = \
                        f'{monthly_stat.stat_data}' \
                        f'<tr id="{row_id}">' \
                        f'<td id="emailer_numerical">{row_number_str}</td>' \
                        f'<td>{subject}<br>{distribution_date}</td>' \
                        f'<td id="emailer_numerical">{sent}</td>' \
                        f'<td id="emailer_numerical">{opens}</td>' \
                        f'<td id="emailer_numerical">{percentage}</td>' \
                        f'</tr>'
                    row_number += 1
                    sent_tally += tracker_data[1]
                    open_tally += tracker_data[0]
            total_opens = f'{open_tally:,}'
            total_sent = f'{sent_tally:,}'
            if open_tally != 0 and sent_tally != 0:
                total_percentage = f'{open_tally / sent_tally:.1%}'
            else:
                total_percentage = '0%'
            monthly_stat.stat_data = \
                f'{monthly_stat.stat_data}' \
                f'<tr id="emailer_title_row">' \
                f'<td><br><br></td>' \
                f'<td id="emailer_numerical">Totals:<br><br></td>' \
                f'<td id="emailer_numerical">{total_sent}<br><br></td>' \
                f'<td id="emailer_numerical">{total_opens}<br><br></td>' \
                f'<td id="emailer_numerical">{total_percentage}<br><br></td>' \
                f'</tr>'
        monthly_stat.save()



from django.db import migrations, models
import django_simple_bulk_emailer.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_simple_bulk_emailer', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscription',
            name='secret_key',
            field=models.CharField(default=django_simple_bulk_emailer.models.create_default_key, max_length=255),
        ),
        migrations.AlterField(
            model_name='bulkemail',
            name='headline',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='bulkemail',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='emailimage',
            name='description',
            field=models.CharField(default='Image', max_length=255, verbose_name='screen reader description'),
        ),
        migrations.AlterField(
            model_name='emailtracker',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='emailtracker',
            name='json_data',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='emailtracker',
            name='subject',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='emailtracker',
            name='subscription_name',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='monthlystat',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='siteprofile',
            name='protocol',
            field=models.CharField(default='https://', max_length=255),
        ),
        migrations.AlterField(
            model_name='subscriber',
            name='first_name',
            field=models.CharField(default='Anonymous', max_length=255),
        ),
        migrations.AlterField(
            model_name='subscriber',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='subscriber',
            name='last_name',
            field=models.CharField(default='Subscriber', max_length=255),
        ),
        migrations.AlterField(
            model_name='subscriber',
            name='subscriber_key',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='associated_model',
            field=models.CharField(default='django_simple_bulk_emailer.models.BulkEmail', max_length=255),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='email_directory',
            field=models.CharField(default='django_simple_bulk_emailer/subscription/emails', max_length=255, verbose_name='email template directory'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='list_name',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='list_slug',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='mc_api',
            field=models.CharField(default='API_key', max_length=255, verbose_name='MailChimp API key'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='mc_list',
            field=models.CharField(default='list_ID', max_length=255, verbose_name='MailChimp audience ID'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='mc_user',
            field=models.CharField(default='username', max_length=255, verbose_name='MailChimp username'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='page_directory',
            field=models.CharField(default='django_simple_bulk_emailer/subscription/pages', max_length=255, verbose_name='page template directory'),
        ),
    ]



from datetime import (
    timedelta,
)
from unittest.mock import (
    patch,
)


from django.contrib.sites.models import (
    Site,
)
from django.test import (
    TestCase,
)
from django.utils import (
    timezone,
)
from django.utils.formats import (
    localize,
)


from django_simple_bulk_emailer.models import (
    BulkEmail,
    EmailDocument,
    EmailImage,
    Subscription,
)


from .functions import (
    attribute_equals,
    attribute_length_equals,
    clear_data_and_files,
    create_email,
    create_site_profile,
    create_stats,
    create_subscription,
    create_subscriber,
    create_tracker,
    fake_now,
    method_output_contains,
    method_output_equals,
    method_output_length_equals,
)


class MixinWrap:
    class BaseMixin(TestCase):
        longMessage = False

        def tearDown(self):
            clear_data_and_files()
            super().tearDown()


class SiteProfileTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.test_instance = create_site_profile()
        super().setUp()

    def test_protocol_domain(self):
        method_output_equals(
            self,
            'protocol_domain',
            'http://127.0.0.1:8000',
        )


class SubscriptionTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.test_instance = create_subscription()
        test_subscriber = create_subscriber()
        test_subscriber.subscriptions.add(self.test_instance)
        super().setUp()

    def test_save(self):
        attribute_equals(
            self,
            {
                'list_slug': 'test-list',
            },
        )

    def test_get_email_class(self):
        method_output_equals(
            self,
            'get_email_class',
            BulkEmail,
        )

    def test_list_link(self):
        method_output_contains(
            self,
            'list_link',
            '<a href="',
        )
        method_output_contains(
            self,
            'list_link',
            '/test-list/" target="_blank">Page</a>',
        )

    def test_subscriber_count(self):
        method_output_equals(
            self,
            'subscriber_count',
            '1',
        )


class SubscriberTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.test_instance = create_subscriber(subscriber_email='Example@Example.com')
        test_subscription_one = create_subscription(list_name='Test Subscription One')
        self.test_instance.subscriptions.add(test_subscription_one)
        test_subscription_two = create_subscription(list_name='Test Subscription Two')
        self.test_instance.subscriptions.add(test_subscription_two)
        super().setUp()

    def test_save(self):
        attribute_length_equals(
            self,
            'subscriber_key',
            64,
        )
        attribute_equals(
            self,
            {
                'subscriber_email': 'example@example.com',
            },
        )

    def test_str(self):
        method_output_equals(
            self,
            '__str__',
            'example@example.com',
        )

    def test_subscription_lists(self):
        method_output_equals(
            self,
            'subscription_lists',
            'Test Subscription One, Test Subscription Two',
        )


class BulkEmailTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        with patch(
            'django_simple_bulk_emailer.models.timezone.now',
            fake_now,
        ):
            self.profile_instance = create_site_profile()
            with self.settings(
                EMAILER_EMAIL_DELETE_DAYS=30,
            ):
                self.test_instance = create_email()
        super().setUp()

    def test_get_deletion_date(self):
        future_time = fake_now() + timedelta(days=30)
        attribute_equals(
            self,
            {
                'deletion_date': future_time.date(),
            },
        )

    def test_short_headline(self):
        method_output_length_equals(
            self,
            'short_headline',
            30,
        )

    def test_short_headline_shorter(self):
        self.test_instance.headline = 'Shorter headline'
        self.test_instance.save()
        method_output_equals(
            self,
            'short_headline',
            'Shorter headline',
        )

    def test_str(self):
        method_output_equals(
            self,
            '__str__',
            'Test headline of many chara...',
        )

    def test_first_paragraph(self):
        method_output_equals(
            self,
            'first_paragraph',
            'Test body text paragraph one.',
        )

    def test_email_subject(self):
        method_output_equals(
            self,
            'email_subject',
            'Test headline of many characters',
        )

    def test_email_subject_updated(self):
        self.test_instance.update_text = 'Test update'
        self.test_instance.save()
        method_output_equals(
            self,
            'email_subject',
            'Updated: Test headline of many characters',
        )

    def test_email_headline(self):
        method_output_equals(
            self,
            'email_headline',
            'Test headline of many characters',
        )

    def test_email_body(self):
        method_output_equals(
            self,
            'email_body',
            '<p>Test body text paragraph one.</p><p>Test body text paragraph two.</p>',
        )

    def test_email_body_substitutions(self):
        with self.settings(
            EMAILER_SUBSTITUTIONS={
                'Test': 'Substitute',
                '</p><p>': '<br><br>',
            },
        ):
            method_output_equals(
                self,
                'email_body',
                '<p>Substitute body text paragraph one.<br><br>Substitute body text paragraph two.</p>',
            )

    def test_email_image(self):
        image_instance = EmailImage.objects.all().first()
        method_output_equals(
            self,
            'email_image',
            image_instance,
        )

    def test_email_documents(self):
        document_instance = EmailDocument.objects.all().first()
        method_output_contains(
            self,
            'email_documents',
            document_instance,
        )

    def test_subscription_name(self):
        subscription_list_name = Subscription.objects.all().first().list_name
        method_output_equals(
            self,
            'subscription_name',
            subscription_list_name,
        )

    def test_subscription_name_unassigned(self):
        self.test_instance.subscription_list = None
        self.test_instance.save()
        method_output_equals(
            self,
            'subscription_name',
            'None chosen',
        )

    def test_headline_slug(self):
        method_output_equals(
            self,
            'headline_slug',
            'test-headline-of-many-characters',
        )

    def test_subscription_url(self):
        method_output_contains(
            self,
            'subscription_url',
            '/test-list/',
        )

    def test_page_url(self):
        method_output_contains(
            self,
            'page_url',
            '/test-list/',
        )
        method_output_contains(
            self,
            'page_url',
            '/test-headline-of-many-characters.html',
        )

    def test_page_preview(self):
        method_output_contains(
            self,
            'page_preview',
            '<a href="',
        )
        method_output_contains(
            self,
            'page_preview',
            '/test-headline-of-many-characters.html" target="_blank">Page</a>',
        )

    def test_email_preview(self):
        method_output_contains(
            self,
            'email_preview',
            '<a href="',
        )
        method_output_contains(
            self,
            'email_preview',
            '/email-preview/test-list/',
        )
        method_output_contains(
            self,
            'email_preview',
            '">Email</a>',
        )

    def test_protocol_domain(self):
        site_instance = Site.objects.get(
            domain=self.profile_instance.domain,
        )
        with self.settings(
            SITE_ID=site_instance.id,
        ):
            method_output_equals(
                self,
                'protocol_domain',
                'http://127.0.0.1:8000',
            )

    def test_static_domain(self):
        site_instance = Site.objects.get(
            domain=self.profile_instance.domain,
        )
        with self.settings(
            STATIC_URL='/static/',
            SITE_ID=site_instance.id,
        ):
            method_output_equals(
                self,
                'static_domain',
                'http://127.0.0.1:8000',
            )

    def test_static_domain_http(self):
        site_instance = Site.objects.get(
            domain=self.profile_instance.domain,
        )
        with self.settings(
            STATIC_URL='http://www.example.com/static/',
            SITE_ID=site_instance.id,
        ):
            method_output_equals(
                self,
                'static_domain',
                '',
            )

    def test_media_domain(self):
        site_instance = Site.objects.get(
            domain=self.profile_instance.domain,
        )
        with self.settings(
            MEDIA_URL='/media/',
            SITE_ID=site_instance.id,
        ):
            method_output_equals(
                self,
                'media_domain',
                'http://127.0.0.1:8000',
            )

    def test_media_domain_http(self):
        site_instance = Site.objects.get(
            domain=self.profile_instance.domain,
        )
        with self.settings(
            MEDIA_URL='http://www.example.com/media/',
            SITE_ID=site_instance.id,
        ):
            method_output_equals(
                self,
                'media_domain',
                '',
            )


class EmailImageTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        email_instance = create_email()
        self.test_instance = email_instance.email_image()
        super().setUp()

    def test_str(self):
        method_output_equals(
            self,
            '__str__',
            'Test description',
        )

    def test_save(self):
        attribute_equals(
            self,
            {
                'output_width': 1080,
            },
        )


class EmailTrackerTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.test_instance = create_tracker()
        super().setUp()

    def test_send_complete_string(self):
        method_output_equals(
            self,
            'send_complete_string',
            localize(timezone.localtime(self.test_instance.send_complete)),
        )


class MonthlyStatTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.test_instance = create_stats()
        super().setUp()

    def test_month_and_year(self):
        method_output_equals(
            self,
            'month_and_year',
            'January 2020',
        )

    def test_str(self):
        method_output_equals(
            self,
            '__str__',
            'January 2020',
        )

    def test_stat_table(self):
        method_output_contains(
            self,
            'stat_table',
            '<table id="emailer_table">',
        )
        method_output_contains(
            self,
            'stat_table',
            'Test stat data',
        )
        method_output_contains(
            self,
            'stat_table',
            '</table>',
        )



from django.db import migrations, models
import django_simple_bulk_emailer.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_simple_bulk_emailer', '0002_auto_20210812_1535'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='bulkemail',
            name='is_updated',
        ),
        migrations.AddField(
            model_name='bulkemail',
            name='secondary_headline',
            field=models.TextField(blank=True, verbose_name='secondary headline (optional)'),
        ),
        migrations.AddField(
            model_name='bulkemail',
            name='update_text',
            field=models.TextField(blank=True, verbose_name='updates (optional)'),
        ),
        migrations.AlterField(
            model_name='bulkemail',
            name='deletion_date',
            field=models.DateField(blank=True, default=django_simple_bulk_emailer.models.get_deletion_date, null=True, verbose_name='deletion date (optional)'),
        ),
    ]



from django.contrib.sites.models import (
    Site,
)
from django.core.management.base import (
    BaseCommand,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone,
)
from django.utils.formats import (
    localize,
)


from ...views import (
    get_universal_email_directory,
    send_email,
)
from ...models import (
    EmailTracker,
    SiteProfile,
    Subscriber,
    Subscription,
)


class Command(BaseCommand):
    help = 'Sends bulk email'

    def handle(self, *args, **options):
        subscriptions = Subscription.objects.order_by(
            'sort_order',
        )
        if subscriptions:
            for subscription in subscriptions:
                email_instance = subscription.get_email_class().objects.filter(
                    sendable=True,
                ).filter(
                    subscription_list=subscription,
                ).order_by(
                    'updated',
                ).first()
                if email_instance:
                    break
            if email_instance:
                ''' Make unavailable to other instances of the function '''
                email_instance.sendable = False
                email_instance.save()
                ''' Create tracker '''
                tracker = EmailTracker.objects.create(
                    subject=email_instance.email_subject(),
                    subscription_name=email_instance.subscription_list.list_name,
                )
                ''' Create email '''
                email_directory = subscription.email_directory
                basic_template = f'{get_universal_email_directory()}/bulk_email_send.html'
                text_template = f'{email_directory}/email_template_text.txt'
                html_template = f'{email_directory}/email_template_html.html'
                site_domain = Site.objects.get_current().domain
                site_profile = SiteProfile.objects.filter(
                    domain=site_domain,
                ).first()
                protocol_domain = site_profile.protocol_domain()
                email_content = {
                    'basic_template': basic_template,
                    'protocol_domain': protocol_domain,
                    'email_instance': email_instance,
                }
                ''' Get subscribers '''
                subscriber_list = Subscriber.objects.filter(
                    subscriptions=subscription,
                )
                ''' Set number_sent at 0 '''
                number_sent = 0
                for subscriber in subscriber_list:
                    ''' Get subscriber-specific information '''
                    tracking_image = reverse(
                        'django_simple_bulk_emailer:opened_email',
                        kwargs={
                            'pk': tracker.pk,
                            'subscriber_key': subscriber.subscriber_key,
                        },
                    )
                    email_content['tracking_image'] = tracking_image
                    to_address = f'"{subscriber.first_name} {subscriber.last_name}" <{subscriber.subscriber_email}>'
                    ''' Send email '''
                    send_email(
                        email_content,
                        list_slug=subscription.list_slug,
                        subscriber_key=subscriber.subscriber_key,
                        text_template=text_template,
                        html_template=html_template,
                        subject=email_instance.email_subject(),
                        to_address=to_address,
                    )
                    ''' Increase number_sent by 1 '''
                    number_sent += 1
                ''' Create send history '''
                send_complete = timezone.now()
                email_instance.send_history = f'<ul>' \
                                              f'<li>Completed: {localize(timezone.localtime(send_complete))}' \
                                              f'<ul>' \
                                              f'<li>Sent to: {email_instance.subscription_list}</li>' \
                                              f'</ul>' \
                                              f'</li>' \
                                              f'</ul>' \
                                              f'{email_instance.send_history}'
                ''' Release email to be sent again '''
                email_instance.sending = False
                email_instance.save()
                ''' Update tracker '''
                tracker.send_complete = send_complete
                tracker.number_sent = number_sent
                tracker.save()



from django import (
    forms,
)
from django.conf import (
    settings,
)
from django.contrib import (
    admin,
)
from django.db import (
    models,
)


from adminsortable2.admin import (
    SortableAdminBase,
    SortableAdminMixin,
    SortableInlineAdminMixin,
)
from django_simple_file_handler.file_types import (
    CHECK_DOC,
    CHECK_WEB_IMAGE,
)
from django_simple_file_handler.validators import (
    CheckExtMIME,
)


from .models import (
    BulkEmail,
    EmailDocument,
    EmailImage,
    MonthlyStat,
    SiteProfile,
    Subscriber,
    Subscription,
)


class BaseAdmin(admin.ModelAdmin):
    actions = None
    readonly_fields = [
        'created',
        'updated',
    ]
    bottom_fieldsets = [
        (
            'Date and time information', {
                'fields': [
                    'created',
                    'updated',
                ],
                'classes': [
                    'collapse',
                ],
            }
        ),
    ]
    fieldsets = bottom_fieldsets
    list_per_page = 20


class SiteProfileAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'protocol',
                        'domain',
                        'name',
                    ]
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    search_fields = [
        'protocol',
        'domain',
        'name',
    ]
    list_display = [
        'name',
        'domain',
    ]
    ordering = [
        'name',
    ]


admin.site.register(
    SiteProfile,
    SiteProfileAdmin,
)


class SubscriptionAdmin(SortableAdminMixin, BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'subscriber_count',
            'secret_key',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'list_name',
                        'descriptive_text',
                        'publicly_visible',
                        'use_pages',
                        'subscriber_count',
                    ]
                }
            ),
            (
                'MailChimp sync', {
                    'fields': [
                        'mc_sync',
                        'mc_user',
                        'mc_api',
                        'mc_list',
                        'secret_key',
                    ],
                    'classes': [
                        'collapse',
                    ]
                }
            ),
            (
                'Advanced settings', {
                    'fields': [
                        'email_directory',
                        'page_directory',
                        'associated_model',
                    ],
                    'classes': [
                        'collapse',
                    ]
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    search_fields = [
        'list_name',
    ]
    list_display = [
        'list_name',
        'subscriber_count',
        'publicly_visible',
        'list_link',
    ]


admin.site.register(
    Subscription,
    SubscriptionAdmin,
)


class SubscriberAdminForm(forms.ModelForm):
    subscriptions = forms.ModelMultipleChoiceField(
        queryset=Subscription.objects.order_by(
            'list_name',
        ),
        label='Subscriptions',
        required=False,
        widget=admin.widgets.FilteredSelectMultiple(
            'subscriptions',
            False,
        )
    )

    class Meta:
        model = Subscriber
        exclude = [
            'subscriber_key',
            'mc_email',
            'mc_synced',
        ]


class SubscriberAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'subscription_lists',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'first_name',
                        'last_name',
                        'subscriber_email',
                        'subscriptions',
                    ]
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    def get_form(self, request, obj=None, **kwargs):
        if obj and not self.has_change_permission(request, obj):
            return super().get_form(request, obj, **kwargs)
        return SubscriberAdminForm

    search_fields = [
        'first_name',
        'last_name',
        'subscriber_email',
    ]
    list_display = [
        'subscriber_email',
        'first_name',
        'last_name',
        'subscription_lists',
    ]
    ordering = [
        'subscriber_email',
    ]


admin.site.register(
    Subscriber,
    SubscriberAdmin,
)


def get_image_widths():
    try:
        width_choices = settings.EMAILER_IMAGE_WIDTHS
    except AttributeError:
        width_choices = [
            (1200, 'Banner'),
            (900, 'Large'),
            (600, 'Medium'),
            (300, 'Small'),
        ]
    return width_choices


class EmailImageInlineForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['saved_file'].validators.append(CheckExtMIME(allowed_attributes=CHECK_WEB_IMAGE))

    image_width = forms.ChoiceField(
        label='Image size',
        choices=get_image_widths(),
    )

    class Meta:
        exclude = []


class EmailImageInline(admin.StackedInline):
    form = EmailImageInlineForm
    model = EmailImage
    fieldsets = [
        (
            None, {
                'fields': [
                    'image_width',
                    'description',
                    'caption',
                    'saved_file',
                ]
            }
        ),
    ]
    formfield_overrides = {
        models.CharField: {
            'widget': forms.TextInput(
                attrs={
                    'size': '95',
                },
            ),
        },
        models.TextField: {
            'widget': forms.Textarea(
                attrs={
                    'rows': 3,
                    'cols': 95,
                },
            ),
        },
    }
    extra = 0
    max_num = 1


class EmailDocumentInlineForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['saved_file'].validators.append(CheckExtMIME(allowed_attributes=CHECK_DOC))

    class Meta:
        exclude = []


class EmailDocumentInline(SortableInlineAdminMixin, admin.TabularInline):
    form = EmailDocumentInlineForm
    model = EmailDocument
    fieldsets = [
        (
            None, {
                'fields': [
                    'title',
                    'extra_text',
                    'saved_file',
                    'sort_order',
                ]
            }
        ),
    ]
    formfield_overrides = {
        models.TextField: {
            'widget': forms.Textarea(
                attrs={
                    'rows': 1,
                },
            ),
        },
    }
    extra = 0


class BulkEmailAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['subscription_list'].queryset = Subscription.objects.filter(
            associated_model__contains=self.instance.__module__,
        ).filter(
            associated_model__contains=self.instance.__class__.__name__,
        )
        self.fields['subscription_list'].empty_label = None

    class Meta:
        model = BulkEmail
        exclude = [
            'sendable',
            'sending',
            'sent',
            'send_history',
        ]
        widgets = {
            'headline': forms.TextInput(
                attrs={
                    'size': '95',
                },
            ),
            'secondary_headline': forms.Textarea(
                attrs={
                    'rows': 3,
                    'cols': 95,
                },
            ),
            'update_text': forms.Textarea(
                attrs={
                    'rows': 3,
                    'cols': 95,
                },
            ),
            'publication_date': admin.widgets.AdminDateWidget,
            'deletion_date': admin.widgets.AdminDateWidget,
        }


class BulkEmailAdmin(SortableAdminBase, BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'subscription_name',
            'short_headline',
            'page_preview',
            'email_preview',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'subscription_list',
                        'headline',
                        'secondary_headline',
                        'update_text',
                        'body_text',
                    ]
                }
            ),
        ]
        self.middle_fieldsets = [
             (
                 None, {
                     'fields': [
                         'published',
                         'publication_date',
                         'deletion_date',
                     ]
                 }
             ),
        ]
        self.fieldsets = self.top_fieldsets + self.middle_fieldsets + self.bottom_fieldsets

    def get_form(self, request, obj=None, **kwargs):
        if obj and not self.has_change_permission(request, obj):
            return super().get_form(request, obj, **kwargs)
        return BulkEmailAdminForm

    inlines = [
        EmailImageInline,
        EmailDocumentInline,
    ]
    search_fields = [
        'headline',
        'body_text',
    ]
    list_display = [
        'short_headline',
        'email_preview',
        'sent',
        'page_preview',
        'published',
        'subscription_name',
        'publication_date',
        'deletion_date',
    ]
    ordering = [
        '-publication_date',
        '-created',
    ]


admin.site.register(
    BulkEmail,
    BulkEmailAdmin,
)


class MonthlyStatAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'month_and_year',
            'stat_table',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'month_and_year',
                        'stat_table',
                    ]
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    list_display = [
        'month_and_year',
    ]
    ordering = [
        '-year_int',
        '-month_int',
    ]

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    class Media:
        css = {
            'all': ('admin/css/django_simple_bulk_emailer.css',)
        }


admin.site.register(
    MonthlyStat,
    MonthlyStatAdmin,
)



from django.apps import (
    AppConfig,
)


class Config(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_simple_bulk_emailer'
    verbose_name = 'Bulk email'

    def ready(self):
        from .signals import (
            handlers,
        )



import hashlib
from random import (
    choice,
)
from string import (
    ascii_lowercase,
    ascii_uppercase,
    digits,
)


from django.core.exceptions import (
    ObjectDoesNotExist,
)
from django.db.models.signals import (
    m2m_changed,
    pre_delete,
    pre_save,
)


from ..models import (
    Subscriber,
    Subscription,
)


def class_set(input_set):
    unchecked_classes = input_set
    unchecked_subclasses = set()
    output_classes = set()
    while len(unchecked_classes) > 0:
        for unchecked_item in unchecked_classes:
            unchecked_subclasses.update(unchecked_item.__subclasses__())
            output_classes.add(unchecked_item)
        unchecked_classes = unchecked_subclasses
        unchecked_subclasses = set()
    return output_classes


def save_sync(sender, instance, *args, **kwargs):
    instance.subscriber_key = ''.join(choice(ascii_lowercase + ascii_uppercase + digits) for _ in range(64))
    instance.subscriber_email = instance.subscriber_email.lower()
    if not instance.mc_email:
        instance.mc_email = instance.subscriber_email
    try:
        saved_object = instance.__class__.objects.get(
            pk=instance.pk,
        )
        changeable_fields = [
            'first_name',
            'last_name',
            'subscriber_email',
        ]
        for field in changeable_fields:
            if getattr(instance, field) != getattr(saved_object, field):
                instance.mc_synced = False
                break
    except ObjectDoesNotExist:
        pass


def m2m_sync(sender, instance, *args, **kwargs):
    instance.mc_synced = False
    instance.save()


def delete_sync(sender, instance, *args, **kwargs):
    subscriber_hash = hashlib.md5(instance.mc_email.encode('utf-8')).hexdigest()
    subscriptions = Subscription.objects.filter(
        mc_sync=True,
    )
    for subscription in subscriptions:
        if subscription in instance.subscriptions.all():
            client = MailChimp(
                mc_api=subscription.mc_api,
                mc_user=subscription.mc_user,
            )
            try:
                client.lists.members.update(
                    list_id=subscription.mc_list,
                    subscriber_hash=subscriber_hash,
                    data={
                        'status': 'unsubscribed',
                    },
                )
            except MailChimpError:
                pass


try:
    from mailchimp3 import MailChimp
    from mailchimp3.mailchimpclient import MailChimpError
    for class_item in class_set(set({Subscriber})):
        pre_save.connect(save_sync, class_item)
        m2m_changed.connect(m2m_sync, class_item.subscriptions.through)
        pre_delete.connect(delete_sync, class_item)
except ImportError:
    pass



from django.test import (
    TestCase,
)


from .functions import (
    check_field_in_form,
    check_form_is_valid,
    clear_data_and_files,
    create_subscription,
)

from django_simple_bulk_emailer.forms import (
    GetSubscriberForm,
    ModifySubscriberForm,
)


class MixinWrap:
    class BaseMixin(TestCase):
        longMessage = False

        def tearDown(self):
            clear_data_and_files()
            super().tearDown()


class GetSubscriberFormTests(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.form_class = GetSubscriberForm
        self.data = {
            'subscriber_email': 'example@example.com',
        }
        super().setUp()

    def test_valid_subscriber_valid_email(self):
        check_form_is_valid(
            self,
            True,
        )

    def test_valid_subscriber_invalid_email(self):
        self.data['subscriber_email'] = 'example_example.com'
        check_form_is_valid(
            self,
            False,
        )

    def test_valid_honeypot_valid_email(self):
        self.data['email'] = 'example@example.com'
        check_form_is_valid(
            self,
            True,
        )

    def test_valid_honeypot_invalid_email(self):
        self.data['email'] = 'example_example.com'
        check_form_is_valid(
            self,
            False,
        )
    '''
    ReCaptcha-related code is not tested
    '''


class SubscriptionSelectorModifySubscriberFormBase(MixinWrap.BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.form_class = ModifySubscriberForm
        self.data = {
            'first_name': 'Test',
            'last_name': 'User',
            'subscriber_email': 'example@example.com',
        }
        super().setUp()


class SubscriptionSelectorTests(SubscriptionSelectorModifySubscriberFormBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_valid_public_subscriptions(self):
        subscription_one = create_subscription(
            list_name='List One',
        )
        subscription_two = create_subscription(
            list_name='List Two',
        )
        self.data['subscription_choices'] = [
            subscription_one.pk,
            subscription_two.pk,
        ]
        check_form_is_valid(
            self,
            True,
        )

    def test_valid_nonpublic_subscription(self):
        subscription_one = create_subscription(
            list_name='List One',
            publicly_visible=False,
        )
        subscription_two = create_subscription(
            list_name='List Two',
        )
        self.data['subscription_choices'] = [
            subscription_one.pk,
            subscription_two.pk,
        ]
        check_form_is_valid(
            self,
            False,
        )


class ModifySubscriberFormTests(SubscriptionSelectorModifySubscriberFormBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.excluded_data = {
            'subscriptions': [1],
            'subscriber_key': 'test_key',
            'mc_email': 'example@example.com',
            'mc_synced': True,
        }
        super().setUp()

    def test_valid_all_required_fields(self):
        check_form_is_valid(
            self,
            True,
        )

    def test_valid_invalid_email(self):
        self.data['subscriber_email'] = 'example_example.com'
        check_form_is_valid(
            self,
            False,
        )

    def test_valid_empty_required_field(self):
        for key, value in self.data.items():
            old_value = value
            self.data[key] = ''
            extra_text = f" — the field being tested was '{key}'"
            check_form_is_valid(
                self,
                False,
                extra_text=extra_text,
            )
            self.data[key] = old_value

    def test_data_excluded_fields(self):
        for key, value in self.excluded_data.items():
            self.data[key] = value
            check_field_in_form(
                self,
                key,
                False,
            )
            del self.data[key]
